# -*- coding: utf-8 -*-

import math

import pytest

from chempy import Reaction
from chempy.units import (
    allclose,
    Backend,
    to_unitless,
    units_library,
    default_units as u,
)
from chempy.util._expr import Log10, Constant
from chempy.util.testing import requires
from ..rates import MassAction, Radiolytic
from .._rates import (
    TPoly,
    RTPoly,
    ShiftedTPoly,
    ShiftedRTPoly,
    ShiftedLog10TPoly,
    TPiecewise,
)


def test_TPolyMassAction():
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, inact_reac={"B": 1})
    p = MassAction(ShiftedTPoly([273.15, 7, 0.2, 0.03, 0.004]))
    res = p({"A": 11, "B": 13, "temperature": 298.15}, reaction=r)
    ref = 7 + 0.2 * 25 + 0.03 * 25 ** 2 + 0.004 * 25 ** 3
    assert abs(res - ref * 13 * 11 ** 2) < 1e-15


def test_ShiftedTPoly_MassAction():
    rate = MassAction(ShiftedTPoly([100, 2, 5, 7]))
    assert rate.args[0].args == [100, 2, 5, 7]
    r = Reaction({"A": 2, "B": 1}, {"P": 1}, rate)
    res = r.rate_expr()({"A": 11, "B": 13, "temperature": 273.15}, reaction=r)
    x = 273.15 - 100
    ref = 11 * 11 * 13 * (2 + 5 * x + 7 * x ** 2)
    assert abs((res - ref) / ref) < 1e-14


@requires(units_library)
def test_ShiftedTPoly__units():
    stp1 = ShiftedTPoly([273.15 * u.kelvin, 5, 7 / u.kelvin])
    allclose(stp1({"temperature": 274.15 * u.kelvin}), 5 + 7)
    allclose(stp1({"temperature": 273.15 * u.kelvin}), 5)
    stp2 = ShiftedTPoly(
        [273.15 * u.kelvin, 5 * u.m, 7 * u.m / u.kelvin, 13 * u.m * u.kelvin ** -2]
    )
    allclose(stp2({"temperature": 274.15 * u.kelvin}), (5 + 7 + 13) * u.m)
    allclose(stp2({"temperature": 273.15 * u.kelvin}), 5 * u.m)


@requires(units_library)
def test_TPolyMassAction__units():
    Mps = u.molar / u.second
    kunit = 1 / u.molar ** 2 / u.second
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, None, {"B": 1})
    p = MassAction(
        ShiftedTPoly(
            [
                273.15 * u.K,
                7 * kunit,
                0.2 * kunit / u.K,
                0.03 * kunit / u.K ** 2,
                0.004 * kunit / u.K ** 3,
            ]
        )
    )
    res = p(
        {"A": 11 * u.molar, "B": 13 * u.molar, "temperature": 298.15 * u.K}, reaction=r
    )
    ref = 7 + 0.2 * 25 + 0.03 * 25 ** 2 + 0.004 * 25 ** 3
    assert abs(res - ref * 13 * 11 ** 2 * Mps) < 1e-15


def test_RTPolyMassAction():
    p = MassAction(ShiftedRTPoly([273.15, 7, 0.2, 0.03, 0.004]))
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, inact_reac={"B": 1})
    res = p({"A": 11, "B": 13, "temperature": 298.15}, reaction=r)
    ref = 7 + 0.2 / 25 + 0.03 / 25 ** 2 + 0.004 / 25 ** 3
    assert abs(res - ref * 13 * 11 ** 2) < 1e-15


@requires(units_library)
def test_RTPolyMassAction__units():
    Mps = u.molar / u.second
    kunit = 1 / u.molar ** 2 / u.second
    p = MassAction(
        ShiftedRTPoly(
            [
                273.15 * u.K,
                7 * kunit,
                0.2 * kunit * u.K,
                0.03 * kunit * u.K ** 2,
                0.004 * kunit * u.K ** 3,
            ]
        )
    )
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, None, {"B": 1})
    res = p(
        {"A": 11 * u.molar, "B": 13 * u.molar, "temperature": 298.15 * u.K}, reaction=r
    )
    ref = 7 + 0.2 / 25 + 0.03 / 25 ** 2 + 0.004 / 25 ** 3
    assert abs(res - ref * 13 * 11 ** 2 * Mps) < 1e-15


def test_Log10TPolyMassAction():
    p = MassAction(10 ** ShiftedTPoly([273.15, 0.7, 0.02, 0.003, 0.0004]))
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, p, {"B": 1})
    res = p({"A": 11, "B": 13, "temperature": 298.15}, reaction=r)
    ref = 10 ** (0.7 + 0.02 * 25 + 0.003 * 25 ** 2 + 0.0004 * 25 ** 3)
    assert abs(res - ref * 13 * 11 ** 2) < 1e-15


@pytest.mark.xfail  # TODO: fails in github-actions, passes locally (on 2 different machines..)
@requires(units_library)
def test_Log10TPolyMassAction__units():
    Mps = u.molar / u.second
    kunit = 1 / u.molar ** 2 / u.second
    p = MassAction(
        Constant(kunit)
        * 10
        ** ShiftedTPoly(
            [273.15 * u.K, 0.7, 0.02 / u.K, 0.003 / u.K ** 2, 0.0004 / u.K ** 3]
        )
    )
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, p, {"B": 1})
    res = p(
        {"A": 11 * u.molar, "B": 13 * u.molar, "temperature": 298.15 * u.K}, reaction=r
    )
    ref = 10 ** (0.7 + 0.02 * 25 + 0.003 * 25 ** 2 + 0.0004 * 25 ** 3)
    assert abs(res - ref * 13 * 11 ** 2 * Mps) < 1e-15


def test_TPolyInLog10MassAction():
    p = MassAction(ShiftedLog10TPoly([2, 0.3, 0.2, 0.03, 0.004]))
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, p, {"B": 1})
    lgT = Log10("temperature")
    lgTref = Log10("Tref")
    res = p(
        {
            "A": 11,
            "B": 13,
            "temperature": 298.15,
            "log10_temperature": lgT,
            "log10_Tref": lgTref,
        },
        reaction=r,
    )
    _T = math.log10(298.15) - 2
    ref = 0.3 + 0.2 * _T + 0.03 * _T ** 2 + 0.004 * _T ** 3
    assert abs(res - ref * 13 * 11 ** 2) < 1e-15


@requires(units_library)
def test_TPolyInLog10MassAction__units():
    Mps = u.molar / u.second
    kunit = 1 / u.molar ** 2 / u.second
    p = MassAction(Constant(kunit) * ShiftedLog10TPoly([2, 0.3, 0.2, 0.03, 0.004]))
    lgT = Log10("temperature" / Constant(u.K))
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, p, {"B": 1})
    res = p(
        {
            "A": 11 * u.molar,
            "B": 13 * u.molar,
            "temperature": 298.15 * u.K,
            "log10_temperature": lgT,
        },
        backend=Backend(),
        reaction=r,
    )
    _T = math.log10(298.15) - 2
    ref = (0.3 + 0.2 * _T + 0.03 * _T ** 2 + 0.004 * _T ** 3) * 13 * 11 ** 2 * Mps
    assert abs(res - ref) < 1e-15 * Mps


def test_TPolyRadiolytic():
    tpr = Radiolytic(ShiftedTPoly([273.15, 1.85e-7, 1e-9]))
    res = tpr({"doserate": 0.15, "density": 0.998, "temperature": 298.15})
    assert abs(res - 0.15 * 0.998 * 2.1e-7) < 1e-15


@requires(units_library)
def test_TPolyRadiolytic__units():
    def _check(r):
        res = r(
            {
                "doserate": 0.15 * u.gray / u.second,
                "density": 0.998 * u.kg / u.decimetre ** 3,
                "temperature": 298.15 * u.K,
            }
        )
        ref = 0.15 * 0.998 * 2.1e-7 * u.molar / u.second
        assert abs(to_unitless((res - ref) / ref)) < 1e-15

    _check(
        Radiolytic(
            ShiftedTPoly(
                [273.15 * u.K, 1.85e-7 * u.mol / u.joule, 1e-9 * u.mol / u.joule / u.K]
            )
        )
    )


def test_TPiecewisePolyMassAction():
    tp1 = TPoly([10, 0.1])
    tp2 = ShiftedTPoly([273.15, 37.315, -0.1])
    pwp = MassAction(TPiecewise([0, tp1, 273.15, tp2, 373.15]))
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, inact_reac={"B": 1})
    res1 = pwp({"A": 11, "B": 13, "temperature": 198.15}, reaction=r)
    ref1 = 11 * 11 * 13 * 29.815
    assert abs((res1 - ref1) / ref1) < 1e-14
    res2 = pwp({"A": 11, "B": 13, "temperature": 298.15}, reaction=r)
    ref2 = 11 * 11 * 13 * (37.315 - 25 * 0.1)
    assert abs((res2 - ref2) / ref2) < 1e-14
    with pytest.raises(ValueError):
        pwp({"A": 11, "B": 13, "temperature": 398.15}, reaction=r)


@requires(units_library)
def test_PiecewiseTPolyMassAction__units():
    tp1 = TPoly(
        [10 / u.molar ** 2 / u.second, 0.1 / u.molar ** 2 / u.second / u.kelvin]
    )
    tp2 = ShiftedTPoly(
        [
            273.15 * u.kelvin,
            37.315 / u.molar ** 2 / u.second,
            -0.1 / u.molar ** 2 / u.second / u.kelvin,
        ]
    )
    pwp = MassAction(
        TPiecewise([0 * u.kelvin, tp1, 273.15 * u.kelvin, tp2, float("inf") * u.kelvin])
    )
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, pwp, {"B": 1})
    res1 = pwp(
        {"A": 11 * u.molar, "B": 13 * u.molar, "temperature": 198.15 * u.kelvin},
        reaction=r,
    )
    ref1 = 11 * 11 * 13 * 29.815 * u.molar / u.second
    assert abs((res1 - ref1) / ref1) < 1e-14


@requires("sympy")
def test_PiecewiseTPolyMassAction__sympy():
    import sympy as sp

    tp1 = TPoly([10, 0.1])
    tp2 = ShiftedTPoly([273.15, 37.315, -0.1])
    pwp = MassAction(TPiecewise([0, tp1, 273.15, tp2, 373.15]))
    T = sp.Symbol("T")
    r = Reaction({"A": 2, "B": 1}, {"C": 1}, inact_reac={"B": 1})
    res1 = pwp({"A": 11, "B": 13, "temperature": T}, backend=sp, reaction=r)
    ref1 = (
        11 ** 2
        * 13
        * sp.Piecewise(
            (10 + 0.1 * T, sp.And(T <= sp.Float(273.15), 0 <= T)),
            (37.315 - 0.1 * (T - 273.15), sp.And(sp.Float(273.15) <= T, T <= sp.Float(373.15))),
            (sp.Symbol("NAN"), True),
        )
    )
    for Tval in [200, 300, 400]:
        assert res1.subs({T: Tval}) == ref1.subs({T: Tval})


def test_Log10PiecewiseRTPolyMassAction():
    p1 = RTPoly([12.281, -3.768e2, -6.673e4, -1.075e7])
    p2 = RTPoly([-47.532, 4.92, -1.036, 0.0])
    ratex = MassAction(10 ** TPiecewise([293.15, p1, 423.15, p2, 623.15]))
    r = Reaction({"e-(aq)": 2}, {"H2": 1, "OH-": 2}, ratex, {"H2O": 2})
    res = ratex({"e-(aq)": 1e-13, "temperature": 293.15}, reaction=r)
    ref = 6.20e9 * 1e-26
    assert abs((res - ref) / ref) < 6e-3


@requires(units_library)
def test_TPiecewise():
    expr0 = ShiftedTPoly([273.15 * u.K, 10 / u.molar / u.s, 0.1 / u.molar / u.s / u.K])
    expr1 = ShiftedTPoly(
        [
            298.15 * u.K,
            12.5 / u.molar / u.s,
            0 / u.molar / u.s / u.K,
            2 / u.molar / u.s / u.K ** 2,
        ]
    )
    pwma = MassAction(
        TPiecewise([273.15 * u.K, expr0, 298.15 * u.K, expr1, 373.15 * u.K])
    )
    r = Reaction({"e-(aq)": 2}, {"H2": 1, "OH-": 2}, inact_reac={"H2O": 2})
    res0 = pwma({"temperature": 293.15 * u.K, "e-(aq)": 1e-13 * u.molar}, reaction=r)
    ref0 = 12 * 1e-26 * u.molar / u.s
    assert allclose(res0, ref0)
    assert not allclose(res0, 2 * ref0)

    res1 = pwma({"temperature": 300.15 * u.K, "e-(aq)": 2e-13 * u.molar}, reaction=r)
    ref1 = 20.5 * 4e-26 * u.molar / u.s
    assert allclose(res1, ref1)
    assert not allclose(res1, ref1 / 2)


@requires(units_library)
def test_MassAction__rate_coeff():
    perMolar_perSecond = u.perMolar_perSecond

    p1 = MassAction(
        Constant(perMolar_perSecond)
        * 10 ** RTPoly([1, 2 * u.kelvin, 3 * u.kelvin ** 2])
    )
    rcoeff1 = p1.rate_coeff({"temperature": 283.15 * u.K})
    ref1 = 10 ** (1 + 2 / 283.15 + 3 / 283.15 ** 2) * perMolar_perSecond
    assert allclose(rcoeff1, ref1)
    rxn1 = Reaction({"A", "B"}, {"C"}, p1)
    rat1 = rxn1.rate({"A": 2, "B": 3, "temperature": 283.15 * u.K})
    assert allclose(rat1["A"], -2 * 3 * ref1)
    assert allclose(rat1["B"], -2 * 3 * ref1)
    assert allclose(rat1["C"], 2 * 3 * ref1)



# -*- coding: utf-8 -*-

from .util.pyutil import ChemPyDeprecationWarning

import warnings

from .electrolytes import (
    A,
    B,
    limiting_log_gamma,
    extended_log_gamma,
    davies_log_gamma,
    limiting_activity_product,
    extended_activity_product,
    davies_activity_product,
)


warnings.warn("use .electrolytes instead of .debye_huckel", ChemPyDeprecationWarning)



# -*- coding: utf-8 -*-

import math

from ..util.pyutil import deprecated
from ..util._expr import Expr


class MassActionEq(Expr):

    argument_names = ("equilibrium_constant",)

    def active_conc_prod(self, variables, backend=math, equilibrium=None):
        result = None
        for exp_factor, stoichs in [(1, equilibrium.prod), (-1, equilibrium.reac)]:
            for k, v in stoichs.items():
                if result is None:
                    result = variables[k] ** (exp_factor * v)
                else:
                    result *= variables[k] ** (exp_factor * v)
        return result

    def eq_const(self, variables, backend=math, **kwargs):
        (eq_c,) = self.all_args(variables, backend=backend, **kwargs)
        return eq_c

    def __call__(self, *args, **kwargs):
        return self.eq_const(*args, **kwargs)

    def equilibrium_equation(self, variables, backend=math, equilibrium=None, **kwargs):
        return self.eq_const(
            variables, backend=backend, **kwargs
        ) - self.active_conc_prod(
            variables, backend=backend, equilibrium=equilibrium, **kwargs
        )

    @classmethod
    def from_callback(cls, callback, attr="eq_const", **kwargs):
        return super(MassActionEq, cls).from_callback(callback, attr=attr, **kwargs)


@deprecated(use_instead=MassActionEq)
class EqExpr(Expr):
    """Baseclass for equilibrium expressions"""

    kw = {"eq": None, "ref": None}


class GibbsEqConst(MassActionEq):
    argument_names = ("dH_over_R", "dS_over_R")
    parameter_keys = ("temperature",)

    def eq_const(self, variables, backend=math, **kwargs):
        dH_over_R, dS_over_R = self.all_args(variables, backend=backend)
        (T,) = self.all_params(variables, backend=backend)
        return backend.exp(dS_over_R - dH_over_R / T)



# -*- coding: utf-8 -*-

import pytest
from pyparsing import ParseException

from ..parsing import (
    formula_to_composition,
    formula_to_html,
    formula_to_latex,
    formula_to_unicode,
    parsing_library,
    to_reaction,
)

from ..testing import requires


@requires(parsing_library)
@pytest.mark.parametrize(
    "species, composition",
    [
        ("H2O*", {1: 2, 8: 1}),
        ("H2O'", {1: 2, 8: 1}),
        ("H2O''", {1: 2, 8: 1}),
        ("H2O'''", {1: 2, 8: 1}),
        ("Na", {11: 1}),
        ("Na*", {11: 1}),
        ("Na'", {11: 1}),
        ("Na''", {11: 1}),
        ("Na'''", {11: 1}),
        ("Na(g)", {11: 1}),
        ("Na*(g)", {11: 1}),
        ("Na'(g)", {11: 1}),
        ("Na''(g)", {11: 1}),
        ("Na'''(g)", {11: 1}),
        ("Na+", {0: 1, 11: 1}),
        ("Na*+", {0: 1, 11: 1}),
        ("Na'+", {0: 1, 11: 1}),
        ("Na''+", {0: 1, 11: 1}),
        ("Na'''+", {0: 1, 11: 1}),
        ("Na+(g)", {0: 1, 11: 1}),
        ("Na*+(g)", {0: 1, 11: 1}),
        ("Na'+(g)", {0: 1, 11: 1}),
        ("Na''+(g)", {0: 1, 11: 1}),
        ("Na'''+(g)", {0: 1, 11: 1}),
        ("Na2CO3*", {11: 2, 6: 1, 8: 3}),
        ("Na2CO3..7H2O*(s)", {11: 2, 6: 1, 8: 10, 1: 14}),
    ],
)
def test_formula_to_composition_primes(species, composition):
    """Should parse special species."""
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species, composition",
    [
        ("CO2(g)", {6: 1, 8: 2}),
        ("CO2(l)", {6: 1, 8: 2}),
        ("CO2(s)", {6: 1, 8: 2}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_state_in_suffixes(species, composition):
    """Should parse species with state in suffixes."""
    assert (
        formula_to_composition(
            species,
            suffixes=("(g)", "(l)", "(s)"),
        )
        == composition
    )


@pytest.mark.parametrize(
    "species, composition",
    [
        ("CO2(aq)", {6: 1, 8: 2}),
        ("H2O(aq)", {1: 2, 8: 1}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_state_not_in_suffixes(species, composition):
    """Should parse species without state in suffixes."""
    assert (
        formula_to_composition(
            species,
            suffixes=("(g)", "(l)", "(s)"),
        )
        == composition
    )


@pytest.mark.parametrize(
    "species, composition",
    [
        ("Li@C60", {3: 1, 6: 60}),
        ("Li@C60Cl", {3: 1, 6: 60, 17: 1}),
        ("(Li@C60)+", {0: 1, 3: 1, 6: 60}),
        ("Na@C60", {11: 1, 6: 60}),
        ("(Na@C60)+", {0: 1, 11: 1, 6: 60}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_caged(species, composition):
    """Should parse cage species."""
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species",
    [
        ("ch3oh"),
        ("Ch3oh"),
        ("Ch3OH"),
        ("ch3OH(l)"),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_fail(species):
    """Should raise an exception."""
    with pytest.raises(ParseException):
        formula_to_composition(species)


@pytest.mark.parametrize(
    "species, composition",
    [
        ("Cl-", {0: -1, 17: 1}),
        ("Fe(SCN)2+", {0: 1, 6: 2, 7: 2, 16: 2, 26: 1}),
        ("Fe(SCN)2+1", {0: 1, 6: 2, 7: 2, 16: 2, 26: 1}),
        ("Fe+3", {0: 3, 26: 1}),
        ("NH4+", {0: 1, 1: 4, 7: 1}),
        ("Na+", {0: 1, 11: 1}),
        ("Na+1", {0: 1, 11: 1}),
        ("OH-", {0: -1, 1: 1, 8: 1}),
        ("SO4-2(aq)", {0: -2, 8: 4, 16: 1}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_ions(species, composition):
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species",
    [
        # Ions.
        ("Cl/-"),
        ("Cl/-(aq)"),
        ("Fe(SCN)2/+"),
        ("Fe(SCN)2/+(aq)"),
        ("Fe/3+"),
        ("Fe/3+(aq)"),
        ("Na/+"),
        ("Na/+(aq)"),
        # Electrons.
        ("e/-"),
        ("e/-(aq)"),
        # Radicals.
        (".NO3/2-"),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_deprecated_charge(species):
    with pytest.raises(ValueError):
        formula_to_composition(species)


@pytest.mark.parametrize(
    "species",
    [
        ("Na+Cl-"),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_bad_charges(species):
    with pytest.raises(ValueError):
        formula_to_composition(species)


@pytest.mark.parametrize(
    "species, composition",
    [
        # With and without water of hydration.
        ("BaCl2", {17: 2, 56: 1}),
        ("BaCl2(s)", {17: 2, 56: 1}),
        ("BaCl2..2H2O(s)", {1: 4, 8: 2, 17: 2, 56: 1}),
        ("Na2CO3..7H2O(s)", {11: 2, 6: 1, 8: 10, 1: 14}),
        ("NaCl", {11: 1, 17: 1}),
        ("NaCl(s)", {11: 1, 17: 1}),
        ("Ni", {28: 1}),
        ("NI", {7: 1, 53: 1}),
        ("KF", {9: 1, 19: 1}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_ionic_compounds(species, composition):
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species, composition",
    [
        # With and without water of hydration.
        ("Al2(SO4)3", {8: 12, 13: 2, 16: 3}),
        ("Al2(SO4)3(s)", {8: 12, 13: 2, 16: 3}),
        ("Al2(SO4)3(aq)", {8: 12, 13: 2, 16: 3}),
        ("K4[Fe(CN)6]", {6: 6, 7: 6, 19: 4, 26: 1}),
        ("K4[Fe(CN)6](s)", {6: 6, 7: 6, 19: 4, 26: 1}),
        ("K4[Fe(CN)6](aq)", {6: 6, 7: 6, 19: 4, 26: 1}),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O",
            {
                1: 50,
                6: 6,
                7: 6,
                8: 25,
                26: 2,
            },
        ),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O(s)",
            {
                1: 50,
                6: 6,
                7: 6,
                8: 25,
                26: 2,
            },
        ),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O(aq)",
            {
                1: 50,
                6: 6,
                7: 6,
                8: 25,
                26: 2,
            },
        ),
        (
            "[Fe(CN)6]-3",
            {
                0: -3,
                6: 6,
                7: 6,
                26: 1,
            },
        ),
        (
            "[Fe(CN)6]-3(aq)",
            {
                0: -3,
                6: 6,
                7: 6,
                26: 1,
            },
        ),
        (
            "Ag[NH3]+",
            {
                0: 1,
                1: 3,
                7: 1,
                47: 1,
            },
        ),
        (
            "[Ni(NH3)6]+2",
            {
                0: 2,
                1: 18,
                7: 6,
                28: 1,
            },
        ),
        (
            "[PtCl6]-2",
            {
                0: -2,
                17: 6,
                78: 1,
            },
        ),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_complexes(species, composition):
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species",
    [
        ("[Fe(CN)6)-3"),
        ("(Fe(CN)6]-3"),
        ("[Fe(CN]6]-3"),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_bad_complexes(species):
    with pytest.raises(ParseException):
        formula_to_composition(species)


@pytest.mark.parametrize(
    "species, composition",
    [
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6",
            {
                6: 6,
                8: 18,
                12: 5.395,
                20: 2.832,
                26: 0.6285,
            },
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6(s)",
            {
                6: 6,
                8: 18,
                12: 5.395,
                20: 2.832,
                26: 0.6285,
            },
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6..8H2O(s)",
            {
                1: 16,
                6: 6,
                8: 26,
                12: 5.395,
                20: 2.832,
                26: 0.6285,
            },
        ),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_fractional_subscripts(species, composition):
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species, composition",
    [
        ("e-", {0: -1}),
        ("e-1", {0: -1}),
        ("e-(aq)", {0: -1}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_solvated_electrons(species, composition):
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species, composition",
    [
        ("H2O", {1: 2, 8: 1}),
        ("((H2O)2OH)12", {1: 60, 8: 36}),
        ("PCl5", {15: 1, 17: 5}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_covalent_compounds(species, composition):
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species, composition",
    [
        ("CH4(g)", {1: 4, 6: 1}),
        ("CH3CH3(g)", {1: 6, 6: 2}),
        # Many ways to write benzene.
        ("C6H6(l)", {1: 6, 6: 6}),
        ("(CH)6(l)", {1: 6, 6: 6}),
        ("CHCHCHCHCHCH(l)", {1: 6, 6: 6}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_organic_compounds(species, composition):
    assert formula_to_composition(species) == composition


@pytest.mark.parametrize(
    "species, composition",
    [
        (".NO2(g)", {7: 1, 8: 2}),
        (".NH2", {1: 2, 7: 1}),
        ("ONOOH", {1: 1, 7: 1, 8: 3}),
        (".ONOO", {7: 1, 8: 3}),
        (".NO3-2", {0: -2, 7: 1, 8: 3}),
        # Deprecated charges as of 0.8.0.
        # (".NO3/2-", {0: -2, 7: 1, 8: 3}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_radicals(species, composition):
    assert formula_to_composition(species) == composition


@requires(parsing_library)
def test_formula_to_composition_structural_formulas():
    with pytest.raises(ValueError):
        formula_to_composition("F-F")


@pytest.mark.parametrize(
    "species, composition",
    [
        ("alpha-FeOOH(s)", {1: 1, 8: 2, 26: 1}),
        ("epsilon-Zn(OH)2(s)", {1: 2, 8: 2, 30: 1}),
    ],
)
@requires(parsing_library)
def test_formula_to_composition_crystal_phases(species, composition):
    assert formula_to_composition(species) == composition


@requires(parsing_library)
def test_to_reaction():
    from chempy.chemistry import Reaction, Equilibrium

    rxn = to_reaction(
        "H+ + OH- -> H2O; 1.4e11; ref={'doi': '10.1039/FT9908601539'}",
        "H+ OH- H2O".split(),
        "->",
        Reaction,
    )
    assert rxn.__class__ == Reaction

    assert rxn.reac["H+"] == 1
    assert rxn.reac["OH-"] == 1
    assert rxn.prod["H2O"] == 1
    assert rxn.param == 1.4e11
    assert rxn.ref["doi"].startswith("10.")

    eq = to_reaction(
        "H+ + OH- = H2O; 1e-14; ref='rt, [H2O] == 1 M'",
        "H+ OH- H2O".split(),
        "=",
        Equilibrium,
    )
    assert eq.__class__ == Equilibrium

    assert eq.reac["H+"] == 1
    assert eq.reac["OH-"] == 1
    assert eq.prod["H2O"] == 1
    assert eq.ref.startswith("rt")

    for s in [
        "2 e-(aq) + (2 H2O) -> H2 + 2 OH- ; 1e6 ; ",
        "2 * e-(aq) + (2 H2O) -> 1 * H2 + 2 * OH- ; 1e6 ; ",
    ]:
        rxn2 = to_reaction(s, "e-(aq) H2 OH- H2O".split(), "->", Reaction)
        assert rxn2.__class__ == Reaction
        assert rxn2.reac["e-(aq)"] == 2
        assert rxn2.inact_reac["H2O"] == 2
        assert rxn2.prod["H2"] == 1
        assert rxn2.prod["OH-"] == 2
        assert rxn2.param == 1e6

    r1 = to_reaction("-> H2O", None, "->", Reaction)
    assert r1.reac == {}
    assert r1.prod == {"H2O": 1}
    assert r1.param is None

    r2 = to_reaction("H2O ->", None, "->", Reaction)
    assert r2.reac == {"H2O": 1}
    assert r2.prod == {}
    assert r2.param is None

    from chempy.kinetics.rates import MassAction

    ma = MassAction([3.14])
    r3 = to_reaction("H+ + OH- -> H2O", None, "->", Reaction, param=ma)
    assert r3.param.args == [3.14]

    rxn3 = to_reaction(
        "H2O + H2O -> H3O+ + OH-", "H3O+ OH- H2O".split(), "->", Reaction
    )
    assert rxn3.reac == {"H2O": 2} and rxn3.prod == {"H3O+": 1, "OH-": 1}

    rxn4 = to_reaction(
        "2 e-(aq) + (2 H2O) + (2 H+) -> H2 + 2 H2O",
        "e-(aq) H2 H2O H+".split(),
        "->",
        Reaction,
    )
    assert (
        rxn4.reac == {"e-(aq)": 2}
        and rxn4.inact_reac == {"H2O": 2, "H+": 2}
        and rxn4.prod == {"H2": 1, "H2O": 2}
    )


@pytest.mark.parametrize(
    "species, latex",
    [
        ("H2O", "H_{2}O"),
        # ("C6H6/+", "C_{6}H_{6}^{+}"),
        ("C6H6+", "C_{6}H_{6}^{+}"),
        # ("C18H38/2+", "C_{18}H_{38}^{2+}"),
        # ("C18H38/+2", "C_{18}H_{38}^{2+}"),
        ("C18H38+2", "C_{18}H_{38}^{2+}"),
        ("NaCl", "NaCl"),
        ("NaCl(s)", "NaCl(s)"),
        ("e-(aq)", "e^{-}(aq)"),
        ("Ca+2(aq)", "Ca^{2+}(aq)"),
        (".NO2(g)", r"^\bullet NO_{2}(g)"),
        (".NH2", r"^\bullet NH_{2}"),
        ("ONOOH", "ONOOH"),
        (".ONOO", r"^\bullet ONOO"),
        # (".NO3/2-", r"^\bullet NO_{3}^{2-}"),
        (".NO3-2", r"^\bullet NO_{3}^{2-}"),
        ("alpha-FeOOH(s)", r"\alpha-FeOOH(s)"),
        ("epsilon-Zn(OH)2(s)", r"\varepsilon-Zn(OH)_{2}(s)"),
        ("Na2CO3..7H2O(s)", r"Na_{2}CO_{3}\cdot 7H_{2}O(s)"),
        ("Na2CO3..1H2O(s)", r"Na_{2}CO_{3}\cdot H_{2}O(s)"),
        ("K4[Fe(CN)6]", r"K_{4}[Fe(CN)_{6}]"),
        ("K4[Fe(CN)6](s)", r"K_{4}[Fe(CN)_{6}](s)"),
        ("K4[Fe(CN)6](aq)", r"K_{4}[Fe(CN)_{6}](aq)"),
        ("[Fe(H2O)6][Fe(CN)6]..19H2O", r"[Fe(H_{2}O)_{6}][Fe(CN)_{6}]\cdot 19H_{2}O"),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O(s)",
            r"[Fe(H_{2}O)_{6}][Fe(CN)_{6}]\cdot 19H_{2}O(s)",
        ),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O(aq)",
            r"[Fe(H_{2}O)_{6}][Fe(CN)_{6}]\cdot 19H_{2}O(aq)",
        ),
        ("[Fe(CN)6]-3", r"[Fe(CN)_{6}]^{3-}"),
        ("[Fe(CN)6]-3(aq)", r"[Fe(CN)_{6}]^{3-}(aq)"),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6",
            r"Ca_{2.832}Fe_{0.6285}Mg_{5.395}(CO_{3})_{6}",
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6(s)",
            r"Ca_{2.832}Fe_{0.6285}Mg_{5.395}(CO_{3})_{6}(s)",
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6..8H2O(s)",
            r"Ca_{2.832}Fe_{0.6285}Mg_{5.395}(CO_{3})_{6}\cdot 8H_{2}O(s)",
        ),
    ],
)
@requires(parsing_library)
def test_formula_to_latex(species, latex):
    assert formula_to_latex(species) == latex


@pytest.mark.parametrize(
    "species, latex",
    [
        # Parentheses.
        ("Fe(CN)6-3", "Fe(CN)_{6}^{3-}"),
        ("((H2O)2OH)12", "((H_{2}O)_{2}OH)_{12}"),
        # Square brackets.
        ("Fe[CN]6-3", "Fe[CN]_{6}^{3-}"),
        ("[(H2O)2OH]12", "[(H_{2}O)_{2}OH]_{12}"),
        # Curly braces.
        ("Fe{CN}6-3", r"Fe\{CN\}_{6}^{3-}"),
        ("{(H2O)2OH}12", r"\{(H_{2}O)_{2}OH\}_{12}"),
    ],
)
@requires(parsing_library)
def test_formula_to_latex_braces(species, latex):
    assert formula_to_latex(species) == latex


@pytest.mark.parametrize(
    "species, latex",
    [
        ("Li@C60", r"Li@C_{60}"),
        ("(Li@C60)+", r"(Li@C_{60})^{+}"),
        ("Na@C60", r"Na@C_{60}"),
        ("(Na@C60)+", r"(Na@C_{60})^{+}"),
    ],
)
@requires(parsing_library)
def test_formula_to_latex_caged(species, latex):
    """Should produce LaTeX for cage species."""
    assert formula_to_latex(species) == latex


@pytest.mark.parametrize(
    "species, unicode",
    [
        ("NH4+", u"NH₄⁺"),
        ("H2O", u"H₂O"),
        # ("C6H6/+", u"C₆H₆⁺"),
        ("C6H6+", u"C₆H₆⁺"),
        # ("Fe(CN)6/3-", u"Fe(CN)₆³⁻"),
        ("Fe(CN)6-3", u"Fe(CN)₆³⁻"),
        # ("C18H38/2+", u"C₁₈H₃₈²⁺"),
        # ("C18H38/+2", u"C₁₈H₃₈²⁺"),
        ("C18H38+2", u"C₁₈H₃₈²⁺"),
        ("((H2O)2OH)12", u"((H₂O)₂OH)₁₂"),
        ("[(H2O)2OH]12", u"[(H₂O)₂OH]₁₂"),
        ("{(H2O)2OH}12", u"{(H₂O)₂OH}₁₂"),
        ("NaCl", u"NaCl"),
        ("NaCl(s)", u"NaCl(s)"),
        ("e-(aq)", u"e⁻(aq)"),
        ("Ca+2(aq)", u"Ca²⁺(aq)"),
        (".NO2(g)", u"⋅NO₂(g)"),
        (".NH2", u"⋅NH₂"),
        ("ONOOH", u"ONOOH"),
        (".ONOO", u"⋅ONOO"),
        # (".NO3/2-", u"⋅NO₃²⁻"),
        (".NO3-2", u"⋅NO₃²⁻"),
        ("alpha-FeOOH(s)", u"α-FeOOH(s)"),
        ("epsilon-Zn(OH)2(s)", u"ε-Zn(OH)₂(s)"),
        ("Na2CO3..7H2O(s)", u"Na₂CO₃·7H₂O(s)"),
        ("Na2CO3..1H2O(s)", u"Na₂CO₃·H₂O(s)"),
        ("K4[Fe(CN)6]", r"K₄[Fe(CN)₆]"),
        ("K4[Fe(CN)6](s)", r"K₄[Fe(CN)₆](s)"),
        ("K4[Fe(CN)6](aq)", r"K₄[Fe(CN)₆](aq)"),
        ("[Fe(H2O)6][Fe(CN)6]..19H2O", r"[Fe(H₂O)₆][Fe(CN)₆]·19H₂O"),
        ("[Fe(H2O)6][Fe(CN)6]..19H2O(s)", r"[Fe(H₂O)₆][Fe(CN)₆]·19H₂O(s)"),
        ("[Fe(H2O)6][Fe(CN)6]..19H2O(aq)", r"[Fe(H₂O)₆][Fe(CN)₆]·19H₂O(aq)"),
        ("[Fe(CN)6]-3", r"[Fe(CN)₆]³⁻"),
        ("[Fe(CN)6]-3(aq)", r"[Fe(CN)₆]³⁻(aq)"),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6",
            r"Ca₂.₈₃₂Fe₀.₆₂₈₅Mg₅.₃₉₅(CO₃)₆",
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6(s)",
            r"Ca₂.₈₃₂Fe₀.₆₂₈₅Mg₅.₃₉₅(CO₃)₆(s)",
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6..8H2O(s)",
            r"Ca₂.₈₃₂Fe₀.₆₂₈₅Mg₅.₃₉₅(CO₃)₆·8H₂O(s)",
        ),
        (
            "Zn(NO3)2..6H2O",
            r"Zn(NO₃)₂·6H₂O",
        ),
    ],
)
@requires(parsing_library)
def test_formula_to_unicode(species, unicode):
    assert formula_to_unicode(species) == unicode


@pytest.mark.parametrize(
    "species, unicode",
    [
        ("Li@C60", r"Li@C₆₀"),
        ("(Li@C60)+", r"(Li@C₆₀)⁺"),
        ("Na@C60", r"Na@C₆₀"),
        ("(Na@C60)+", r"(Na@C₆₀)⁺"),
    ],
)
@requires(parsing_library)
def test_formula_to_unicode_caged(species, unicode):
    """Should produce LaTeX for cage species."""
    assert formula_to_unicode(species) == unicode


@pytest.mark.parametrize(
    "species, html",
    [
        ("H2O", "H<sub>2</sub>O"),
        # ("C6H6/+", "C<sub>6</sub>H<sub>6</sub><sup>+</sup>"),
        ("C6H6+", "C<sub>6</sub>H<sub>6</sub><sup>+</sup>"),
        # ("Fe(CN)6/3-", "Fe(CN)<sub>6</sub><sup>3-</sup>"),
        ("Fe(CN)6-3", "Fe(CN)<sub>6</sub><sup>3-</sup>"),
        # ("C18H38/2+", "C<sub>18</sub>H<sub>38</sub><sup>2+</sup>"),
        # ("C18H38/+2", "C<sub>18</sub>H<sub>38</sub><sup>2+</sup>"),
        ("C18H38+2", "C<sub>18</sub>H<sub>38</sub><sup>2+</sup>"),
        ("((H2O)2OH)12", "((H<sub>2</sub>O)<sub>2</sub>OH)<sub>12</sub>"),
        ("[(H2O)2OH]12", "[(H<sub>2</sub>O)<sub>2</sub>OH]<sub>12</sub>"),
        ("{(H2O)2OH}12", "{(H<sub>2</sub>O)<sub>2</sub>OH}<sub>12</sub>"),
        ("NaCl", "NaCl"),
        ("NaCl(s)", "NaCl(s)"),
        ("e-(aq)", "e<sup>-</sup>(aq)"),
        ("Ca+2(aq)", "Ca<sup>2+</sup>(aq)"),
        (".NO2(g)", r"&sdot;NO<sub>2</sub>(g)"),
        (".NH2", r"&sdot;NH<sub>2</sub>"),
        ("ONOOH", "ONOOH"),
        (".ONOO", r"&sdot;ONOO"),
        # (".NO3/2-", r"&sdot;NO<sub>3</sub><sup>2-</sup>"),
        (".NO3-2", r"&sdot;NO<sub>3</sub><sup>2-</sup>"),
        ("alpha-FeOOH(s)", r"&alpha;-FeOOH(s)"),
        ("epsilon-Zn(OH)2(s)", (r"&epsilon;-Zn(OH)<sub>2</sub>(s)")),
        ("Na2CO3..7H2O(s)", "Na<sub>2</sub>CO<sub>3</sub>&sdot;7H<sub>2</sub>O(s)"),
        ("Na2CO3..1H2O(s)", "Na<sub>2</sub>CO<sub>3</sub>&sdot;H<sub>2</sub>O(s)"),
        ("K4[Fe(CN)6]", r"K<sub>4</sub>[Fe(CN)<sub>6</sub>]"),
        ("K4[Fe(CN)6](s)", r"K<sub>4</sub>[Fe(CN)<sub>6</sub>](s)"),
        ("K4[Fe(CN)6](aq)", r"K<sub>4</sub>[Fe(CN)<sub>6</sub>](aq)"),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O",
            r"[Fe(H<sub>2</sub>O)<sub>6</sub>][Fe(CN)<sub>6</sub>]&sdot;19H<sub>2</sub>O",
        ),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O(s)",
            r"[Fe(H<sub>2</sub>O)<sub>6</sub>][Fe(CN)<sub>6</sub>]&sdot;19H<sub>2</sub>O(s)",
        ),
        (
            "[Fe(H2O)6][Fe(CN)6]..19H2O(aq)",
            r"[Fe(H<sub>2</sub>O)<sub>6</sub>][Fe(CN)<sub>6</sub>]&sdot;19H<sub>2</sub>O(aq)",
        ),
        ("[Fe(CN)6]-3", r"[Fe(CN)<sub>6</sub>]<sup>3-</sup>"),
        ("[Fe(CN)6]-3(aq)", r"[Fe(CN)<sub>6</sub>]<sup>3-</sup>(aq)"),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6",
            r"Ca<sub>2.832</sub>Fe<sub>0.6285</sub>Mg<sub>5.395</sub>(CO<sub>3</sub>)<sub>6</sub>",
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6(s)",
            r"Ca<sub>2.832</sub>Fe<sub>0.6285</sub>Mg<sub>5.395</sub>(CO<sub>3</sub>)<sub>6</sub>(s)",
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6..8H2O(s)",
            r"Ca<sub>2.832</sub>Fe<sub>0.6285</sub>Mg<sub>5.395</sub>(CO<sub>3</sub>)<sub>6</sub>&sdot;8H<sub>2</sub>O(s)",
        ),
        (
            "Ca2.832Fe0.6285Mg5.395(CO3)6..8H2O(s)",
            r"Ca<sub>2.832</sub>Fe<sub>0.6285</sub>Mg<sub>5.395</sub>(CO<sub>3</sub>)<sub>6</sub>&sdot;8H<sub>2</sub>O(s)",
        ),
        (
            "Zn(NO3)2..6H2O",
            r"Zn(NO<sub>3</sub>)<sub>2</sub>&sdot;6H<sub>2</sub>O",
        ),
    ],
)
@requires(parsing_library)
def test_formula_to_html(species, html):
    assert formula_to_html(species) == html


@pytest.mark.parametrize(
    "species, html",
    [
        ("Li@C60", r"Li@C<sub>60</sub>"),
        ("(Li@C60)+", r"(Li@C<sub>60</sub>)<sup>+</sup>"),
        ("Na@C60", r"Na@C<sub>60</sub>"),
        ("(Na@C60)+", r"(Na@C<sub>60</sub>)<sup>+</sup>"),
    ],
)
@requires(parsing_library)
def test_formula_to_html_caged(species, html):
    """Should produce HTML for cage species."""
    assert formula_to_html(species) == html


def test_composition_dot_as_crystal_water_chempy08x():
    ref = {30: 1, 7: 2, 8: 12, 1: 12}
    assert formula_to_composition('Zn(NO3)2{}6H2O'.format('\u00B7')) == ref
    assert formula_to_composition('Zn(NO3)2..6H2O') == ref



# -*- coding: utf-8 -*-

from itertools import chain

import pytest

from ..util.testing import requires
from ..util.parsing import parsing_library
from ..units import default_units, units_library, allclose
from ..chemistry import Substance, Reaction
from ..reactionsystem import ReactionSystem


@requires(parsing_library, "numpy")
def test_ReactionSystem():
    import numpy as np

    kw = dict(substance_factory=Substance.from_formula)
    r1 = Reaction.from_string("H2O -> H+ + OH-", "H2O H+ OH-", name="r1")
    rs = ReactionSystem([r1], "H2O H+ OH-", **kw)
    r2 = Reaction.from_string("H2O -> 2 H+ + OH-", "H2O H+ OH-", name="r2")
    with pytest.raises(ValueError):
        ReactionSystem([r2], "H2O H+ OH-", **kw)
    with pytest.raises(ValueError):
        ReactionSystem([r1, r1], "H2O H+ OH-", **kw)
    assert rs.as_substance_index("H2O") == 0
    assert rs.as_substance_index(0) == 0
    varied, varied_keys = rs.per_substance_varied(
        {"H2O": 55.4, "H+": 1e-7, "OH-": 1e-7},
        {"H+": [1e-8, 1e-9, 1e-10, 1e-11], "OH-": [1e-3, 1e-2]},
    )
    assert varied_keys == ("H+", "OH-")
    assert len(varied.shape) == 3
    assert varied.shape[:-1] == (4, 2)
    assert varied.shape[-1] == 3
    assert np.all(varied[..., 0] == 55.4)
    assert np.all(varied[:, 1, 2] == 1e-2)

    assert rs["r1"] is r1
    rs.rxns.append(r2)
    assert rs["r2"] is r2
    with pytest.raises(KeyError):
        rs["r3"]
    rs.rxns.append(Reaction({}, {}, 0, name="r2", checks=()))
    with pytest.raises(ValueError):
        rs["r2"]

    empty_rs = ReactionSystem([])
    rs2 = empty_rs + rs
    assert rs2 == rs
    rs3 = rs + empty_rs
    assert rs3 == rs


@requires(parsing_library)
def test_ReactionSystem__missing_substances_from_keys():
    r1 = Reaction({"H2O"}, {"H+", "OH-"})
    with pytest.raises(ValueError):
        ReactionSystem([r1], substances={"H2O": Substance.from_formula("H2O")})
    kw = dict(
        missing_substances_from_keys=True, substance_factory=Substance.from_formula
    )
    rs = ReactionSystem([r1], substances={"H2O": Substance.from_formula("H2O")}, **kw)
    assert rs.substances["OH-"].composition == {0: -1, 1: 1, 8: 1}


@requires(parsing_library)
def test_ReactionSystem__check_balance():
    rs1 = ReactionSystem.from_string(
        "\n".join(["2 NH3 -> N2 + 3 H2", "N2H4 -> N2 + 2 H2"])
    )
    assert rs1.check_balance(strict=True)
    rs2 = ReactionSystem.from_string(
        "\n".join(["2 A -> B", "B -> 2A"]), substance_factory=Substance
    )
    assert not rs2.check_balance(strict=True)
    assert rs2.composition_balance_vectors() == ([], [])


def test_ReactionSystem__per_reaction_effect_on_substance():
    rs = ReactionSystem([Reaction({"H2": 2, "O2": 1}, {"H2O": 2})])
    assert rs.per_reaction_effect_on_substance("H2") == {0: -2}
    assert rs.per_reaction_effect_on_substance("O2") == {0: -1}
    assert rs.per_reaction_effect_on_substance("H2O") == {0: 2}


def test_ReactionSystem__rates():
    rs = ReactionSystem([Reaction({"H2O"}, {"H+", "OH-"}, 11)])
    assert rs.rates({"H2O": 3, "H+": 5, "OH-": 7}) == {
        "H2O": -11 * 3,
        "H+": 11 * 3,
        "OH-": 11 * 3,
    }


def test_ReactionSystem__rates__cstr():
    k = 11
    rs = ReactionSystem([Reaction({"H2O2": 2}, {"O2": 1, "H2O": 2}, k)])
    c0 = {"H2O2": 3, "O2": 5, "H2O": 53}
    fr = 7
    fc = {"H2O2": 13, "O2": 17, "H2O": 23}
    r = k * c0["H2O2"] ** 2
    ref = {
        "H2O2": -2 * r + fr * fc["H2O2"] - fr * c0["H2O2"],
        "O2": r + fr * fc["O2"] - fr * c0["O2"],
        "H2O": 2 * r + fr * fc["H2O"] - fr * c0["H2O"],
    }
    variables = dict(
        chain(c0.items(), [("fc_" + key, val) for key, val in fc.items()], [("fr", fr)])
    )
    assert (
        rs.rates(variables, cstr_fr_fc=("fr", {sk: "fc_" + sk for sk in rs.substances}))
        == ref
    )


@requires("numpy")
def test_ReactionSystem__html_tables():
    r1 = Reaction({"A": 2}, {"A"}, name="R1")
    r2 = Reaction({"A"}, {"A": 2}, name="R2")
    rs = ReactionSystem([r1, r2])
    ut, unc = rs.unimolecular_html_table()
    assert unc == {0}
    from chempy.printing import html

    assert (
        html(ut, with_name=False)
        == u'<table><tr><td>A</td><td ><a title="1: A → 2 A">R2</a></td></tr></table>'
    )

    bt, bnc = rs.bimolecular_html_table()
    assert bnc == {1}
    assert html(bt, with_name=False) == (
        u'<table><th></th><th>A</th>\n<tr><td>A</td><td ><a title="0: 2 A → A">R1</a></td></tr></table>'
    )


@requires(parsing_library, "numpy")
def test_ReactionSystem__substance_factory():
    r1 = Reaction.from_string("H2O -> H+ + OH-", "H2O H+ OH-")
    rs = ReactionSystem([r1], "H2O H+ OH-", substance_factory=Substance.from_formula)
    assert rs.net_stoichs(["H2O"]) == [-1]
    assert rs.net_stoichs(["H+"]) == [1]
    assert rs.net_stoichs(["OH-"]) == [1]
    assert rs.substances["H2O"].composition[8] == 1
    assert rs.substances["OH-"].composition[0] == -1
    assert rs.substances["H+"].charge == 1


@requires(units_library)
def test_ReactionSystem__as_per_substance_array_dict():
    mol = default_units.mol
    m = default_units.metre
    M = default_units.molar
    rs = ReactionSystem([], [Substance("H2O")])
    c = rs.as_per_substance_array({"H2O": 1 * M}, unit=M)
    assert c.dimensionality == M.dimensionality
    assert abs(c[0] / (1000 * mol / m ** 3) - 1) < 1e-16

    c = rs.as_per_substance_array({"H2O": 1})
    with pytest.raises(KeyError):
        c = rs.as_per_substance_array({"H": 1})

    assert rs.as_per_substance_dict([42]) == {"H2O": 42}


@requires(parsing_library)
def test_ReactionSystem__add():
    rs1 = ReactionSystem.from_string(
        "\n".join(["2 H2O2 -> O2 + 2 H2O", "H2 + O2 -> H2O2"])
    )
    rs2 = ReactionSystem.from_string("\n".join(["2 NH3 -> N2 + 3 H2"]))
    rs3 = rs1 + rs2
    assert rs1 == rs1
    assert rs1 != rs2
    assert rs3 != rs1
    assert len(rs1.rxns) == 2 and len(rs2.rxns) == 1 and len(rs3.rxns) == 3
    for k in "H2O2 O2 H2O H2 NH3 N2".split():
        assert k in rs3.substances
    rs1 += rs2
    assert len(rs1.rxns) == 3 and len(rs2.rxns) == 1
    assert rs1 == rs3

    rs4 = ReactionSystem.from_string("H2O -> H+ + OH-; 1e-4")
    rs4 += [Reaction({"H+", "OH-"}, {"H2O"}, 1e10)]
    assert len(rs4.rxns) == 2
    assert rs4.rxns[0].reac == {"H2O": 1}
    assert rs4.rxns[1].reac == {"H+": 1, "OH-": 1}
    res = rs4.rates({"H2O": 1, "H+": 1e-7, "OH-": 1e-7})
    for k in "H2O H+ OH-".split():
        assert abs(res[k]) < 1e-16

    rs5 = ReactionSystem.from_string("H3O+ -> H+ + H2O")
    rs6 = rs4 + rs5
    rs7 = rs6 + (Reaction.from_string("H+ + H2O -> H3O+"),)
    assert len(rs7.rxns) == 4

    with pytest.raises(ValueError):
        rs5 += (rs1, rs2)

    with pytest.raises(ValueError):
        rs5 + (rs1, rs2)

    rs1 = ReactionSystem.from_string("O2 + H2 -> H2O2")
    rs1.substances["H2O2"].data["D"] = 123
    rs2 = ReactionSystem.from_string("H2O2 -> 2 OH")
    rs2.substances["H2O2"].data["D"] = 456
    rs2.substances["OH"].data["D"] = 789
    rs3 = rs2 + rs1
    assert (
        rs3.substances["H2O2"].data["D"] == 123
        and rs3.substances["OH"].data["D"] == 789
    )
    assert rs3.rxns[0].reac == {"H2O2": 1}
    assert rs3.rxns[1].reac == {"O2": 1, "H2": 1}
    assert len(rs3.rxns) == 2
    rs2 += rs1
    assert (
        rs2.substances["H2O2"].data["D"] == 123
        and rs2.substances["OH"].data["D"] == 789
    )
    assert rs2.rxns[0].reac == {"H2O2": 1}
    assert rs2.rxns[1].reac == {"O2": 1, "H2": 1}
    assert len(rs2.rxns) == 2


@requires(parsing_library, units_library)
def test_ReactionSystem__from_string():
    rs = ReactionSystem.from_string("-> H + OH; Radiolytic(2.1e-7)", checks=())
    assert rs.rxns[0].reac == {}
    assert rs.rxns[0].prod == {"H": 1, "OH": 1}
    assert rs.rxns[0].param.args == [2.1e-7]
    ref = 2.1e-7 * 0.15 * 998
    assert rs.rates({"doserate": 0.15, "density": 998}) == {"H": ref, "OH": ref}

    (r2,) = ReactionSystem.from_string("H2O + H2O + H+ -> H3O+ + H2O").rxns
    assert r2.reac == {"H2O": 2, "H+": 1}
    assert r2.prod == {"H2O": 1, "H3O+": 1}

    rs2 = ReactionSystem.from_string(
        """
 #  H2O -> OH + H
  H+ + OH- -> H2O; 4*pi*(4e-9*m**2/s + 2e-9*m**2/s)*0.44*nm*Avogadro_constant; ref='made up #hashtag'  # comment
#H+ + OH- -> H2O
"""
    )
    assert len(rs2.rxns) == 1
    assert sorted(rs2.substances.keys()) == sorted("H2O H+ OH-".split())
    assert allclose(
        rs2.rxns[0].param, 1.99786e10 / default_units.M / default_units.s, rtol=1e-5
    )
    assert rs2.rxns[0].ref == "made up #hashtag"


@requires(parsing_library, "numpy")
def test_ReactionSystem__from_string__symbolics():
    rs3 = ReactionSystem.from_string(
        """
A -> B; 'kA'
B -> C; 0
""",
        substance_factory=Substance,
    )
    rs3.rxns[1].param = 2 * rs3.rxns[0].param
    assert rs3.rates(dict(A=29, B=31, kA=42)) == {
        "A": -29 * 42,
        "B": 29 * 42 - 2 * 31 * 42,
        "C": 2 * 31 * 42,
    }


@requires(parsing_library, units_library)
def test_ReactionSystem__from_string__units():
    (r3,) = ReactionSystem.from_string(
        "(H2O) -> e-(aq) + H+ + OH; Radiolytic(2.1e-7*mol/J)"
    ).rxns
    assert len(r3.reac) == 0 and r3.inact_reac == {"H2O": 1}
    assert r3.prod == {"e-(aq)": 1, "H+": 1, "OH": 1}
    from chempy.kinetics.rates import Radiolytic

    mol, J = default_units.mol, default_units.J
    assert r3.param == Radiolytic(2.1e-7 * mol / J)
    assert r3.param != Radiolytic(2.0e-7 * mol / J)
    assert r3.param != Radiolytic(2.1e-7)
    assert r3.order() == 0

    k = 1e-4 / default_units.second
    rs = ReactionSystem.from_string(
        """
    H2O -> H+ + OH-; {}
    """.format(
            repr(k)
        )
    )
    assert allclose(rs.rxns[0].param, k)


@requires(parsing_library, "numpy")
def test_ReactionSystem__from_string___special_naming():
    rs = ReactionSystem.from_string(
        """
H2O* + H2O -> 2 H2O
H2O* -> OH + H
"""
    )  # excited water
    for sk in "H2O* H2O OH H".split():
        assert sk in rs.substances
    assert rs.substances["H2O*"].composition == {1: 2, 8: 1}
    assert rs.categorize_substances() == dict(
        accumulated={"OH", "H", "H2O"},
        depleted={"H2O*"},
        unaffected=set(),
        nonparticipating=set(),
    )


@requires(parsing_library)
def test_ReactionSystem__from_string__string_rate_const():
    rsys = ReactionSystem.from_string("H+ + OH- -> H2O; 'kf'")
    (r2,) = rsys.rxns
    assert r2.reac == {"OH-": 1, "H+": 1}
    assert r2.prod == {"H2O": 1}
    r2str = r2.string(rsys.substances, with_param=True)
    assert r2str.endswith("; 'kf'")


@requires("numpy")
def test_ReactionSystem__upper_conc_bounds():
    rs = ReactionSystem.from_string(
        "\n".join(["2 NH3 -> N2 + 3 H2", "N2H4 -> N2 +   2  H2"])
    )
    c0 = {"NH3": 5, "N2": 7, "H2": 11, "N2H4": 2}
    _N = 5 + 14 + 4
    _H = 15 + 22 + 8
    ref = {
        "NH3": min(_N, _H / 3),
        "N2": _N / 2,
        "H2": _H / 2,
        "N2H4": min(_N / 2, _H / 4),
    }
    res = rs.as_per_substance_dict(rs.upper_conc_bounds(c0))
    assert res == ref


@requires("numpy")
def test_ReactionSystem__upper_conc_bounds__a_substance_no_composition():
    rs = ReactionSystem.from_string(
        """
    H2O -> e-(aq) + H2O+
    H2O+ + e-(aq) -> H2O
    """
    )
    c0 = {"H2O": 55.0, "e-(aq)": 2e-3, "H2O+": 3e-3}
    _O = 55 + 3e-3
    _H = 2 * 55 + 2 * 3e-3
    ref = {
        "H2O": min(_O, _H / 2),
        "e-(aq)": float("inf"),
        "H2O+": min(_O, _H / 2),
    }
    res = rs.as_per_substance_dict(rs.upper_conc_bounds(c0))
    assert res == ref


@requires(parsing_library)
def test_ReactionSystem__identify_equilibria():
    rsys = ReactionSystem.from_string(
        """
    2 H2 +  O2 -> 2 H2O     ; 1e-3
           H2O -> H+ + OH-  ; 1e-4/55.35
      H+ + OH- -> H2O       ; 1e10
         2 H2O -> 2 H2 + O2
    """
    )
    assert rsys.identify_equilibria() == [(0, 3), (1, 2)]


@requires(parsing_library, "numpy")
def test_ReactionSystem__categorize_substances():
    rsys1 = ReactionSystem.from_string(
        """
    2 H2 +  O2 -> 2 H2O     ; 1e-3
           H2O -> H+ + OH-  ; 1e-4/55.35
      H+ + OH- -> H2O       ; 1e10
         2 H2O -> 2 H2 + O2
    """
    )
    assert all(not s for s in rsys1.categorize_substances().values())

    rsys2 = ReactionSystem.from_string(
        "\n".join(["2 NH3 -> N2 + 3 H2", "N2H4 -> N2 +   2  H2"])
    )
    assert rsys2.categorize_substances() == dict(
        accumulated={"N2", "H2"},
        depleted={"NH3", "N2H4"},
        unaffected=set(),
        nonparticipating=set(),
    )

    rsys3 = ReactionSystem.from_string("H+ + OH- -> H2O; 'kf'")
    assert rsys3.categorize_substances() == dict(
        accumulated={"H2O"},
        depleted={"H+", "OH-"},
        unaffected=set(),
        nonparticipating=set(),
    )

    rsys4 = ReactionSystem(
        [Reaction({"H2": 2, "O2": 1}, {"H2O": 2})], "H2 O2 H2O N2 Ar"
    )
    assert rsys4.categorize_substances() == dict(
        accumulated={"H2O"},
        depleted={"H2", "O2"},
        unaffected=set(),
        nonparticipating={"N2", "Ar"},
    )

    rsys5 = ReactionSystem.from_string(
        """
    A -> B; MassAction(unique_keys=('k1',))
    B + C -> A + C; MassAction(unique_keys=('k2',))
    2 B -> B + C; MassAction(unique_keys=('k3',))
    """,
        substance_factory=lambda formula: Substance(formula),
    )
    assert rsys5.categorize_substances() == dict(
        accumulated={"C"}, depleted=set(), unaffected=set(), nonparticipating=set()
    )

    rsys6 = ReactionSystem.from_string("""H2O2 + Fe+3 + (H2O2) -> 2 H2O + O2 + Fe+3""")
    assert rsys6.rxns[0].order() == 2  # the additional H2O2 within parenthesis
    assert rsys6.categorize_substances() == dict(
        accumulated={"H2O", "O2"},
        depleted={"H2O2"},
        unaffected={"Fe+3"},
        nonparticipating=set(),
    )


@requires(parsing_library, "numpy")
def test_ReactionSystem__split():
    a = """
    2 H2 +  O2 -> 2 H2O     ; 1e-3
           H2O -> H+ + OH-  ; 1e-4/55.35
      H+ + OH- -> H2O       ; 1e10
        2 H2O  -> 2 H2 + O2"""
    b = """
        2 N    -> N2"""
    c = """
        2 ClBr -> Cl2 + Br2
    """
    rsys1 = ReactionSystem.from_string(a + b + c)
    res = rsys1.split()
    ref = list(map(ReactionSystem.from_string, [a, b, c]))
    for rs in chain(res, ref):
        rs.sort_substances_inplace()
    res1a, res1b, res1c = res
    ref1a, ref1b, ref1c = ref
    assert res1a == ref1a
    assert res1b == ref1b
    assert res1c == ref1c
    assert res1c != ref1a
    assert rsys1.categorize_substances() == dict(
        accumulated={"N2", "Cl2", "Br2"},
        depleted={"N", "ClBr"},
        unaffected=set(),
        nonparticipating=set(),
    )


def test_ReactionSystem__subset():
    r1 = Reaction({"NH3": 2}, {"N2": 1, "H2": 3})
    r2 = Reaction({"N2H4": 1}, {"N2": 1, "H2": 2})
    rs1 = ReactionSystem([r1, r2])
    rs2, rs3 = rs1.subset(lambda r: "N2H4" in r.keys())
    assert len(rs1.rxns) == 2 and len(rs2.rxns) == 1
    assert rs2 == ReactionSystem([r2])
    assert rs3 == ReactionSystem([r1])


@requires(parsing_library)
def test_ReactionSystem__concatenate():
    rs1 = ReactionSystem.from_string(
        """
    H + OH -> H2O; 1e10; name='rs1a'
    2 H2O2 -> 2 H2O + O2; 1e-7; name='rs1b'
"""
    )
    rs2 = ReactionSystem.from_string(
        """
    H + OH -> H2O; 1e11; name='rs2a'
    H2O2 -> H2 + O2; 1e-9; name='rs2b'
"""
    )
    rs, skipped = ReactionSystem.concatenate([rs1, rs2])
    (sr,) = skipped.rxns
    assert sr.name == "rs2a"
    assert rs.rxns[-1].name == "rs2b"



# -*- coding: utf-8 -*-
import re


try:
    import numpy as np
except ImportError:
    np = None


# See https://github.com/python-quantities/python-quantities/pull/112
def format_units_html(udict, font="%s", mult=r"&sdot;", paren=False):
    """
    Replace the units string provided with an equivalent html string.

    Exponentiation (m**2) will be replaced with superscripts (m<sup>2</sup>})

    No formatting is done, change `font` argument to e.g.:
    '<span style="color: #0000a0">%s</span>' to have text be colored blue.

    Multiplication (*) are replaced with the symbol specified by the mult
    argument. By default this is the latex &sdot; symbol.  Other useful options
    may be '' or '*'.

    If paren=True, encapsulate the string in '(' and ')'

    """
    from quantities.markup import format_units

    res = format_units(udict)
    if res.startswith("(") and res.endswith(")"):
        # Compound Unit
        compound = True
    else:
        # Not a compound unit
        compound = False
    # Replace exponentiation (**exp) with ^{exp}
    res = re.sub(r"\*{2,2}(?P<exp>\d+)", r"<sup>\g<exp></sup>", res)
    # Remove multiplication signs
    res = re.sub(r"\*", mult, res)
    if paren and not compound:
        res = "(%s)" % res
    res = font % res
    return res


def _patch_pow0_py35(pq):
    try:
        pq.metre ** 0
    except Exception:
        pq.quantity.Quantity.__pow__ = pq.quantity.check_uniform(
            lambda self, other: np.power(self, other)
        )


def _patch_quantities(pq):
    # See https://github.com/python-quantities/python-quantities/pull/112
    if not hasattr(pq.dimensionality.Dimensionality, "html"):
        pq.dimensionality.Dimensionality.html = property(
            lambda self: format_units_html(self)
        )

    # See https://github.com/python-quantities/python-quantities/pull/116
    a = pq.UncertainQuantity([1, 2], pq.m, [0.1, 0.2])
    if (-a).uncertainty[0] != a.uncertainty[0]:
        pq.UncertainQuantity.__neg__ = lambda self: self * -1
    a = pq.UncertainQuantity([1, 2], pq.m, [0.1, 0.2])
    assert (-a).uncertainty[0] == (a * -1).uncertainty[0]

    # See https://github.com/python-quantities/python-quantities/pull/126
    _patch_pow0_py35(pq)
    assert (3 * pq.m) ** 0 == 1 * pq.dimensionless



# -*- coding: utf-8 -*-
""" Functions for chemical formulae and reactions """


from collections import defaultdict

import re

from .pyutil import memoize
from .periodic import symbols

parsing_library = "pyparsing"  # info used for selective testing.


def get_parsing_context():
    """returns the default dictionary for parsing strings in chempy"""
    import chempy
    from chempy.kinetics import rates
    from chempy.units import default_units, default_constants, to_unitless

    globals_ = dict(to_unitless=to_unitless, chempy=chempy)

    def _update(mod, keys=None):
        if keys is None:
            keys = dir(mod)
        globals_.update({k: getattr(mod, k) for k in keys if not k.startswith("_")})

    try:
        import numpy
    except ImportError:

        def _numpy_not_installed_raise(*args, **kwargs):
            raise ImportError("numpy not installed, no such method")

        class numpy:
            array = staticmethod(_numpy_not_installed_raise)
            log = staticmethod(_numpy_not_installed_raise)
            exp = staticmethod(_numpy_not_installed_raise)

    _update(numpy, keys="array log exp".split())  # could of course add more
    _update(rates)
    _update(chempy)
    for df in [default_units, default_constants]:
        if df is not None:
            globals_.update(df.as_dict())
    return globals_


@memoize()
def _get_formula_parser():
    """Create a forward pyparsing parser for chemical formulae

    BNF for simple chemical formula (no nesting)

        integer :: '0'..'9'+
        element :: 'A'..'Z' 'a'..'z'*
        term :: element [integer]
        formula :: term+


    BNF for nested chemical formula

        integer :: '0'..'9'+
        element :: 'A'..'Z' 'a'..'z'*
        term :: (element | '(' formula ')') [integer]
        formula :: term+

    Notes
    -----
    The code in this function is from an answer on StackOverflow:
        http://stackoverflow.com/a/18555142/790973
        written by:
            Paul McGuire, http://stackoverflow.com/users/165216/paul-mcguire
        in answer to the question formulated by:
            Thales MG, http://stackoverflow.com/users/2708711/thales-mg
        the code is licensed under 'CC-WIKI'.
        (see: http://blog.stackoverflow.com/2009/06/attribution-required/)

    Documentation for the desired product.  Original documentation
    above.

    Create a chemical formula parser.

    Parse a chemical formula, including elements, nested ions,
    complexes, charges (ions), hydrates, and state symbols.

    BNF for nested chemical formula with complexes

        count :: ( '1'..'9'? | '1'..'9'' '0'..'9'+ )
        element :: 'A'..'Z' 'a'..'z'*
        charge :: ( '-' | '+' ) ( '1'..'9'? | '1'..'9'' '0'..'9'+ )
        prime :: ( "*" | "'" )*
        term :: (element
                 | '(' formula ')'
                 | '{' formula '}'
                 | '[' formula ']' ) count prime charge?
        formula :: term+
        hydrate :: ( '..' | '\u00B7' ) count? formula
        state :: '(' ( 's' | 'l' | 'g' | 'aq' | 'cr' ) ')'
        compound :: count formula hydrate? state?

    Parse a chemical formula, including elements, non-integer
    subscripts, nested ions, complexes, charges (ions), hydrates, and
    state symbols.

    BNF for nested chemical formula with complexes

        count :: ( '1'..'9'? | '1'..'9'' '0'..'9'+ | '0'..'9'+ '.' '0'..'9'+ )
        element :: 'A'..'Z' 'a'..'z'*
        charge :: ( '-' | '+' ) ( '1'..'9'? | '1'..'9'' '0'..'9'+ )
        prime :: ( "*" | "'" )*
        term :: (element
                 | '(' formula ')'
                 | '{' formula '}'
                 | '[' formula ']' ) count prime charge?
        formula :: term+
        hydrate :: ( '..' | '\u00B7' ) count? formula
        state :: '(' ( 's' | 'l' | 'g' | 'aq' | 'cr' ) ')'
        compound :: count formula hydrate? state?
    """
    _p = __import__(parsing_library)
    Forward, Group, OneOrMore = _p.Forward, _p.Group, _p.OneOrMore
    Optional, ParseResults, Regex = _p.Optional, _p.ParseResults, _p.Regex
    Suppress = _p.Suppress

    # Define and suppress the grouping symbols.
    LCB = Suppress(Regex(r"\{"))
    RCB = Suppress(Regex(r"\}"))
    LSB = Suppress(Regex(r"\["))
    RSB = Suppress(Regex(r"\]"))
    LP = Suppress(Regex(r"\("))
    RP = Suppress(Regex(r"\)"))

    # Define and suppress the caged symbol.
    caged = Suppress(Regex(r"\@"))

    # Primes/stars for marking special species in reactions.
    primes = Suppress(Regex(r"[*']+"))

    # Parse counts (subscripts and coefficients).
    count = Regex(r"(\d+\.\d+|\d*)")
    count.setParseAction(lambda t: 1 if t[0] == "" else float(t[0]))

    # Parse states.
    state = Suppress(Regex(r"\((s|l|g|aq|cr)\)"))

    # Elements, 1-118, official symbols.
    element = Regex(
        r"A[cglmrstu]"
        "|B[aehikr]?"
        "|C[adeflmnorsu]?"
        "|D[bsy]"
        "|E[rsu]"
        "|F[elmr]?"
        "|G[ade]"
        "|H[efgos]?"
        "|I[nr]?"
        "|Kr?"
        "|L[airuv]"
        "|M[cdgnot]"
        "|N[abdehiop]?"
        "|O[gs]?"
        "|P[abdmortu]?"
        "|R[abefghnu]"
        "|S[bcegimnr]?"
        "|T[abcehilms]"
        "|U"
        "|V"
        "|W"
        "|Xe"
        "|Yb?"
        "|Z[nr]"
    ).setResultsName("element", listAllMatches=True)

    # forward declare 'formula' so it can be used in definition of 'term'
    formula = Forward()

    term = Group(
        (
            element
            | Group(LP + formula + RP)("subgroup")
            | Group(LSB + formula + RSB)("subgroup")
            | Group(LCB + formula + RCB)("subgroup")
            | Group(caged + formula)("subgroup")
        )
        + Optional(count, default=1)("mult")
        + Optional(state)("state")
        + Optional(primes)("primes")
    )

    # add parse actions for parse-time processing

    # parse action to multiply out subgroups
    def multiplyContents(tokens):
        t = tokens[0]
        # if these tokens contain a subgroup, then use multiplier to
        # extend counts of all elements in the subgroup
        if t.subgroup:
            mult = t.mult
            for term in t.subgroup:
                term[1] *= mult
            return t.subgroup

    term.setParseAction(multiplyContents)

    # add parse action to sum up multiple references to the same element
    def sumByElement(tokens):
        elementsList = [t[0] for t in tokens]

        # construct set to see if there are duplicates
        duplicates = len(elementsList) > len(set(elementsList))

        # if there are duplicate element names, sum up by element and
        # return a new nested ParseResults
        if duplicates:
            ctr = defaultdict(int)
            for t in tokens:
                ctr[t[0]] += t[1]
            return ParseResults([ParseResults([k, v]) for k, v in ctr.items()])

    # define contents of a formula as one or more terms
    formula << OneOrMore(term)
    formula.setParseAction(sumByElement)

    return formula


def _get_charge(chgstr):

    if chgstr == "+":
        return 1
    elif chgstr == "-":
        return -1

    for token, anti, sign in zip("+-", "-+", (1, -1)):
        if token in chgstr:
            if anti in chgstr:
                raise ValueError("Invalid charge description (+ & - present)")

            before, after = chgstr.split(token)

            if len(before) > 0 and len(after) > 0:
                raise ValueError("Values both before and after charge token")

            if len(after) > 0:
                return sign * int(1 if after == "" else after)

    raise ValueError("Invalid charge description (+ or - missing)")


def _formula_to_parts(formula, prefixes, suffixes):
    # Drop prefixes and suffixes.
    drop_pref, drop_suff = [], []
    for ign in prefixes:
        if formula.startswith(ign):
            drop_pref.append(ign)
            formula = formula[len(ign) :]
    for ign in suffixes:
        if formula.endswith(ign):
            drop_suff.append(ign)
            formula = formula[: -len(ign)]

    # Extract charge.
    if "/" in formula:
        raise ValueError(
            "Slashes ('/') in charge strings are deprecated."
            "  Use `Fe+3` instead of `Fe/3+`."
        )
    else:
        for token in "+-":
            if token in formula:
                if formula.count(token) > 1:
                    raise ValueError("Multiple tokens: %s" % token)
                parts = formula.split(token)
                parts[1] = token + parts[1]
                break
        else:
            parts = [formula, None]

    return parts + [tuple(drop_pref), tuple(drop_suff[::-1])]


def _parse_stoich(stoich):
    # Special case:  the electron is not an element.
    if stoich == "e":
        return {}

    comp = {}
    for k, n in _get_formula_parser().parseString(stoich, parseAll=True):
        # Only use rational subscripts if necessary as
        # ``sympy.linsolve()`` does not like non-integers when
        # balancing reactions.
        if n == int(n):
            comp[symbols.index(k) + 1] = int(n)
        else:
            comp[symbols.index(k) + 1] = n

    return comp


_greek_letters = (
    "alpha",
    "beta",
    "gamma",
    "delta",
    "epsilon",
    "zeta",
    "eta",
    "theta",
    "iota",
    "kappa",
    "lambda",
    "mu",
    "nu",
    "xi",
    "omicron",
    "pi",
    "rho",
    "sigma",
    "tau",
    "upsilon",
    "phi",
    "chi",
    "psi",
    "omega",
)
_greek_u = "αβγδεζηθικλμνξοπρστυφχψω"

_latex_mapping = {k + "-": "\\" + k + "-" for k in _greek_letters}
_latex_mapping["epsilon-"] = "\\varepsilon-"
_latex_mapping["omicron-"] = "o-"
_latex_mapping["."] = "^\\bullet "
_latex_infix_mapping = {"..": "\\cdot "}

_unicode_mapping = {k + "-": v + "-" for k, v in zip(_greek_letters, _greek_u)}
_unicode_mapping["."] = "⋅"
_unicode_infix_mapping = {"..": "\u00b7"}  # 0x00b7: '·'

_html_mapping = {k + "-": "&" + k + ";-" for k in _greek_letters}
_html_mapping["."] = "&sdot;"
# _html_infix_mapping = _html_mapping
_html_infix_mapping = {"..": "&sdot;"}


def _get_leading_integer(s):
    m = re.findall(r"^\d+", s)
    if len(m) == 0:
        m = 1
    elif len(m) == 1:
        s = s[len(m[0]) :]
        m = int(m[0])
    else:
        raise ValueError("Failed to parse: %s" % s)
    return m, s


def formula_to_composition(
    formula, prefixes=None, suffixes=("(s)", "(l)", "(g)", "(aq)")
):
    """Parse composition of formula representing a chemical formula

    Composition is represented as a dict mapping int -> int (atomic
    number -> multiplicity). "Atomic number" 0 represents net charge.

    Parameters
    ----------
    formula: str
        Chemical formula, e.g. 'H2O', 'Fe+3', 'Cl-'
    prefixes: iterable strings
        Prefixes to ignore, e.g. ('.', 'alpha-')
    suffixes: tuple of strings
        Suffixes to ignore, e.g. ('(g)', '(s)')

    Examples
    --------
    >>> formula_to_composition('NH4+') == {0: 1, 1: 4, 7: 1}
    True
    >>> formula_to_composition('.NHO-(aq)') == {0: -1, 1: 1, 7: 1, 8: 1}
    True
    >>> formula_to_composition('Na2CO3..7H2O') == {11: 2, 6: 1, 8: 10, 1: 14}
    True
    >>> formula_to_composition('UO2.3') == {92: 1, 8: 2.3}
    True

    """
    if prefixes is None:
        prefixes = _latex_mapping.keys()

    stoich_tok, chg_tok = _formula_to_parts(formula, prefixes, suffixes)[:2]
    tot_comp = {}
    if '\u00b7' in stoich_tok:
        parts = stoich_tok.split('\u00b7')
    else:
        parts = stoich_tok.split("..")

    for idx, stoich in enumerate(parts):
        if idx == 0:
            m = 1
        else:
            m, stoich = _get_leading_integer(stoich)
        comp = _parse_stoich(stoich)
        for k, v in comp.items():
            if k not in tot_comp:
                tot_comp[k] = m * v
            else:
                tot_comp[k] += m * v

    if chg_tok is not None:
        tot_comp[0] = _get_charge(chg_tok)

    return tot_comp


def _subs(string, patterns):
    for patt, repl in patterns.items():
        string = string.replace(patt, repl)

    return string


def _parse_multiplicity(strings, substance_keys=None):
    """
    Examples
    --------
    >>> _parse_multiplicity(['2 H2O2', 'O2']) == {'H2O2': 2, 'O2': 1}
    True
    >>> _parse_multiplicity(['2 * H2O2', 'O2']) == {'H2O2': 2, 'O2': 1}
    True
    >>> _parse_multiplicity(['']) == {}
    True
    >>> _parse_multiplicity(['H2O', 'H2O']) == {'H2O': 2}
    True

    """
    result = {}
    for items in [re.split(" \\* | ", s) for s in strings]:
        items = [x for x in items if x != ""]
        if len(items) == 0:
            continue
        elif len(items) == 1:
            if items[0] not in result:
                result[items[0]] = 0
            result[items[0]] += 1
        elif len(items) == 2:
            if items[1] not in result:
                result[items[1]] = 0
            result[items[1]] += (
                float(items[0]) if "." in items[0] or "e" in items[0] else int(items[0])
            )
        else:
            raise ValueError("To many parts in substring")
    if substance_keys is not None:
        for k in result:
            if k not in substance_keys:
                raise ValueError("Unknown substance_key: %s" % k)
    return result


def to_reaction(line, substance_keys, token, Cls, globals_=None, **kwargs):
    """Parses a string into a Reaction object and substances

    Reac1 + 2 Reac2 + (2 Reac1) -> Prod1 + Prod2; 10**3.7; ref='doi:12/ab'
    Reac1 = Prod1; 2.1;

    Parameters
    ----------
    line: str
        string representation to be parsed
    substance_keys: iterable of strings
        Allowed names, e.g. ('H2O', 'H+', 'OH-')
    token : str
        delimiter token between reactant and product side
    Cls : class
        e.g. subclass of Reaction
    globals_: dict (optional)
        Globals passed on to :func:`eval`, when ``None``:
        `chempy.units.default_units` is used with 'chempy'
        and 'default_units' extra entries.

    Notes
    -----
    This function calls :func:`eval`, hence there are severe security concerns
    with running this on untrusted data.

    """
    if globals_ is None:
        globals_ = get_parsing_context()
    parts = line.rstrip("\n").split(";")
    stoich = parts[0].strip()
    if len(parts) > 2:
        kwargs.update(eval("dict(" + ";".join(parts[2:]) + "\n)", globals_ or {}))
    if len(parts) > 1:
        param = parts[1].strip()
    else:
        param = kwargs.pop("param", "None")

    if isinstance(param, str):
        if param.startswith("'") and param.endswith("'") and "'" not in param[1:-1]:
            from ..kinetics.rates import MassAction
            from ._expr import Symbol

            param = MassAction(Symbol(unique_keys=(param[1:-1],)))
        else:
            param = None if globals_ is False else eval(param, globals_)

    if token not in stoich:
        raise ValueError("Missing token: %s" % token)

    reac_prod = [[y.strip() for y in x.split(" + ")] for x in stoich.split(token)]

    act, inact = [], []
    for elements in reac_prod:
        act.append(
            _parse_multiplicity(
                [x for x in elements if not x.startswith("(")], substance_keys
            )
        )
        inact.append(
            _parse_multiplicity(
                [x[1:-1] for x in elements if x.startswith("(") and x.endswith(")")],
                substance_keys,
            )
        )

    # stoich coeff -> dict
    return Cls(
        act[0], act[1], param, inact_reac=inact[0], inact_prod=inact[1], **kwargs
    )


def _formula_to_format(
    sub,
    sup,
    formula,
    prefixes=None,
    infixes=None,
    suffixes=("(s)", "(l)", "(g)", "(aq)"),
):
    parts = _formula_to_parts(formula, prefixes.keys(), suffixes)
    if '\u00b7' in parts[0]:
        stoichs = parts[0].split('\u00b7')
    else:
        stoichs = parts[0].split("..")
    string = ""
    for idx, stoich in enumerate(stoichs):
        if idx == 0:
            m = 1
        else:
            m, stoich = _get_leading_integer(stoich)
            string += _subs("..", infixes)
        if m != 1:
            string += str(m)
        string += re.sub(r"([0-9]+\.[0-9]+|[0-9]+)", lambda m: sub(m.group(1)), stoich)

    if parts[1] is not None:
        chg = _get_charge(parts[1])
        if chg < 0:
            token = "-" if chg == -1 else "%d-" % -chg
        if chg > 0:
            token = "+" if chg == 1 else "%d+" % chg
        string += sup(token)
    if len(parts) > 4:
        raise ValueError("Incorrect formula")
    pre_str = "".join(map(lambda x: _subs(x, prefixes), parts[2]))
    return pre_str + string + "".join(parts[3])


def formula_to_latex(formula, prefixes=None, infixes=None, **kwargs):
    r"""Convert formula string to latex representation

    Parameters
    ----------
    formula: str
        Chemical formula, e.g. 'H2O', 'Fe+3', 'Cl-'
    prefixes: dict
        Prefix transformations, default: greek letters and .
    infixes: dict
        Infix transformations, default: .
    suffixes: iterable of str
        What suffixes not to interpret, default: (s), (l), (g), (aq)

    Examples
    --------
    >>> formula_to_latex('NH4+')
    'NH_{4}^{+}'
    >>> formula_to_latex('Fe(CN)6+2')
    'Fe(CN)_{6}^{2+}'
    >>> formula_to_latex('Fe(CN)6+2(aq)')
    'Fe(CN)_{6}^{2+}(aq)'
    >>> formula_to_latex('.NHO-(aq)')
    '^\\bullet NHO^{-}(aq)'
    >>> formula_to_latex('alpha-FeOOH(s)')
    '\\alpha-FeOOH(s)'

    """
    if prefixes is None:
        prefixes = _latex_mapping
    if infixes is None:
        infixes = _latex_infix_mapping
    return _formula_to_format(
        lambda x: "_{%s}" % x,
        lambda x: "^{%s}" % x,
        # formula,
        re.sub(r"([{}])", r"\\\1", formula) if re.search(r"[{}]", formula) else formula,
        prefixes,
        infixes,
        **kwargs
    )


_unicode_sub = {
    ".": ".",
}

for k, v in enumerate("₀₁₂₃₄₅₆₇₈₉."):
    _unicode_sub[str(k)] = v

_unicode_sup = {
    "+": "⁺",
    "-": "⁻",
}

for k, v in enumerate("⁰¹²³⁴⁵⁶⁷⁸⁹"):
    _unicode_sup[str(k)] = v


def formula_to_unicode(formula, prefixes=None, infixes=None, **kwargs):
    """Convert formula string to unicode string representation

    Parameters
    ----------
    formula : str
        Chemical formula, e.g. 'H2O', 'Fe+3', 'Cl-'
    prefixes : dict
        Prefix transofmrations, default: greek letters and .
    infixes : dict
        Infix transofmrations, default: .
    suffixes : tuple of strings
        Suffixes to keep, e.g. ('(g)', '(s)')

    Examples
    --------
    >>> formula_to_unicode('NH4+') == u'NH₄⁺'
    True
    >>> formula_to_unicode('Fe(CN)6+2') == u'Fe(CN)₆²⁺'
    True
    >>> formula_to_unicode('Fe(CN)6+2(aq)') == u'Fe(CN)₆²⁺(aq)'
    True
    >>> formula_to_unicode('.NHO-(aq)') == u'⋅NHO⁻(aq)'
    True
    >>> formula_to_unicode('alpha-FeOOH(s)') == u'α-FeOOH(s)'
    True

    """
    if prefixes is None:
        prefixes = _unicode_mapping
    if infixes is None:
        infixes = _unicode_infix_mapping
    return _formula_to_format(
        lambda x: "".join(_unicode_sub[str(_)] for _ in x),
        lambda x: "".join(_unicode_sup[str(_)] for _ in x),
        formula,
        prefixes,
        infixes,
        **kwargs
    )


def formula_to_html(formula, prefixes=None, infixes=None, **kwargs):
    """Convert formula string to html string representation

    Parameters
    ----------
    formula : str
        Chemical formula, e.g. 'H2O', 'Fe+3', 'Cl-'
    prefixes : dict
        Prefix transformations, default: greek letters and .
    infixes : dict
        Infix transformations, default: .
    suffixes : tuple of strings
        Suffixes to keep, e.g. ('(g)', '(s)')

    Examples
    --------
    >>> formula_to_html('NH4+')
    'NH<sub>4</sub><sup>+</sup>'
    >>> formula_to_html('Fe(CN)6+2')
    'Fe(CN)<sub>6</sub><sup>2+</sup>'
    >>> formula_to_html('Fe(CN)6+2(aq)')
    'Fe(CN)<sub>6</sub><sup>2+</sup>(aq)'
    >>> formula_to_html('.NHO-(aq)')
    '&sdot;NHO<sup>-</sup>(aq)'
    >>> formula_to_html('alpha-FeOOH(s)')
    '&alpha;-FeOOH(s)'

    """
    if prefixes is None:
        prefixes = _html_mapping
    if infixes is None:
        infixes = _html_infix_mapping
    return _formula_to_format(
        lambda x: "<sub>%s</sub>" % x,
        lambda x: "<sup>%s</sup>" % x,
        formula,
        prefixes,
        infixes,
        **kwargs
    )



# -*- coding: utf-8 -*-


def electrical_mobility_from_D(D, charge, T, constants=None, units=None):
    """
    Calculates the electrical mobility through Einstein-Smoluchowski relation.

    Parameters
    ----------
    D: float with unit
        Diffusion coefficient
    charge: integer
        charge of the species
    T: float with unit
        Absolute temperature
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin and b0 = 1 mol/kg
        else:
            see source code for what attributes are used.
            Tip: pass quantities.constants
    units: object (optional, default: None)
        attributes accessed: meter, Kelvin and mol

    Returns
    -------
    Electrical mobility

    """

    if constants is None:
        kB = 1.38064852e-23
        e = 1.60217662e-19
        if units is not None:
            kB *= units.joule / units.kelvin / units.mol
            e *= units.coulomb
    else:
        kB = constants.Boltzmann_constant
        e = constants.elementary_charge
    return D * charge * e / (kB * T)



# -*- coding: utf-8 -*-


import math

from ..eyring import eyring_equation, EyringParam, EyringParamWithUnits
from chempy.util.testing import requires
from chempy.units import allclose, units_library, default_units as u


_kB_over_h = 2.083664399411865234375e10
_R = 8.314472


def test_eyring_equation():
    dH, dS = 40e3, 1e2
    T = 123.45
    ref = _kB_over_h * T * math.exp(dS / _R) * math.exp(-dH / _R / T)
    res = eyring_equation(dH, dS, T)
    assert abs((res - ref) / ref) < 1e-10


def test_EyringParam():
    T = 273.15
    k = EyringParam(40e3, 1e2)(T)
    ref = _kB_over_h * T * math.exp(1e2 / _R) * math.exp(-40e3 / _R / T)
    assert abs((k - ref) / ref) < 1e-7


@requires(units_library)
def test_EyringParamWithUnits():
    mol = u.mol
    J = u.joule
    K = u.kelvin
    T = 273.15 * K
    k = EyringParamWithUnits(42e3 * J / mol, 123 * J / mol / K)(T)
    ref = _kB_over_h * 273.15 * math.exp(123 / _R) * math.exp(-42e3 / _R / 273.15)
    ref /= u.second
    assert allclose(k, ref)



import json
from .web import CSSPrinter, _html_clsname

_js_rsys_template = """
var cls_names_substances = %(cls_names_substances)s;
var substance_row_cls_irrel = %(substance_row_cls_irrel)s;
var elms = {};
var n = {}, nsubstances = cls_names_substances.length;
var nirrel = {};
function changeColor(classname, color) {
    var curN = n[classname];
    for(var i = 0; i < curN; i++) {
        elms[classname][i].style.backgroundColor = color;
    }
}
function toggleVisibility(classname_substance) {
    var curN = nirrel[classname_substance];
    for (var i=0; i<curN; i++) {
        var objs = document.getElementsByClassName(substance_row_cls_irrel[classname_substance][i]);
        for (var j=0; j<objs.length; ++j){
            objs[j].style.display = objs[j].style.display == "none" ? "table-row" : "none";
        }
    }
}
function resetTab(tab){
    tab.style.border = "0px";
    var rows = tab.getElementsByTagName('tr');
    [].forEach.call(rows, function(row){
        row.style.display = "table-row";
    });
    tab.getElementsByTagName('th')[0].innerHTML = tab.ori_header +
        "<br>(click on species to show a subset of reactions)";
};

for(var k = 0; k < nsubstances; k++) {
    var curClass = cls_names_substances[k];
    var curIrrel = substance_row_cls_irrel[k];
    elms[curClass] = document.getElementsByClassName(curClass);
    n[curClass] = elms[curClass].length;
    nirrel[curClass] = substance_row_cls_irrel[curClass].length;
    var curN = n[curClass];
    for(var i = 0; i < curN; i++) {
        elms[curClass][i].onmouseover = function() {
            changeColor(this.className, "LightBlue");
        };
        elms[curClass][i].onmouseout = function() {
            changeColor(this.className, "inherit");
        };
        elms[curClass][i].onclick = function() {
            var tab = this.closest("table");
            resetTab(tab);
            tab.style.border = "1px dashed #000000";
            toggleVisibility(this.className);
            tab.getElementsByTagName('th')[0].innerHTML = tab.ori_header +
                 "<br>Only showing reactions involving: " + this.innerHTML +
                 " (double-click to reset)";
        };
    }
};
var chempy_tabs = document.querySelectorAll('table.chempy_%(rsys_id)d');
[].forEach.call(chempy_tabs, function(tab){
    tab.ori_header = tab.getElementsByTagName('th')[0].innerHTML;
    tab.ondblclick = function(){
        resetTab(this);
        this.scrollIntoView();
    };
});
[].forEach.call(chempy_tabs, function(tab){
    resetTab(tab);
});
"""


def _js_rsys(cls_names_substances, substance_row_cls_irrel, rsys_id):
    # from https://stackoverflow.com/a/12786869/790973
    if cls_names_substances is None:
        return ""
    return _js_rsys_template % dict(
        cls_names_substances=json.dumps(cls_names_substances),
        substance_row_cls_irrel=json.dumps(substance_row_cls_irrel),
        rsys_id=rsys_id,
    )


class JSPrinter(CSSPrinter):
    """Prints javascript-enabled HTML representations"""

    def _print_ReactionSystem(self, rsys, **kwargs):
        tab = super(JSPrinter, self)._print_ReactionSystem(rsys, **kwargs)
        _script_tag = '<script type="text/javascript">%s</script>'
        substances = self._get("substances", **kwargs)
        cls_names_substances = list(map(_html_clsname, substances))
        return tab + _script_tag % _js_rsys(
            cls_names_substances=cls_names_substances,
            substance_row_cls_irrel={
                cns: [  # reactions not involving sk
                    self._tr_id(rsys, i)
                    for i in range(rsys.nr)
                    if i not in rsys.substance_participation(sk)
                ]
                for cns, sk in zip(cls_names_substances, substances)
            },
            rsys_id=id(rsys),
        )


def javascript(obj, **settings):
    return JSPrinter(settings).doprint(obj)



# Lint-only option from https://stackoverflow.com/a/60045845/12968623.
def pytest_collection_modifyitems(session, config, items):
    if config.getoption("--lint-only"):
        lint_items = []
        for linter in ["black", "flake8"]:
            if config.getoption(f"--{linter}"):
                lint_items.extend(
                    [item for item in items if item.get_closest_marker(linter)]
                )
        items[:] = lint_items


def pytest_addoption(parser):
    parser.addoption(
        "--lint-only",
        action="store_true",
        default=False,
        help="Run linting checks only.",
    )



# -*- coding: utf-8 -*-

from collections import defaultdict
from functools import reduce
from operator import mul

try:
    import numpy as np
except ImportError:
    np = None

import pytest

from chempy import ReactionSystem
from chempy.units import (
    allclose,
    logspace_from_lin,
    unitless_in_registry,
    to_unitless,
    SI_base_registry,
    rescale,
    default_units as u,
)
from chempy.util.testing import requires
from ..integrated import binary_rev
from ..ode import get_odesys
from .._native import get_native


decay_analytic = {
    0: lambda y0, k, t: (y0[0] * np.exp(-k[0] * t)),
    1: lambda y0, k, t: (
        y0[1] * np.exp(-k[1] * t)
        + y0[0] * k[0] / (k[1] - k[0]) * (np.exp(-k[0] * t) - np.exp(-k[1] * t))
    ),
    2: lambda y0, k, t: (
        y0[2] * np.exp(-k[2] * t)
        + y0[1] * k[1] / (k[2] - k[1]) * (np.exp(-k[1] * t) - np.exp(-k[2] * t))
        + k[1]
        * k[0]
        * y0[0]
        / (k[1] - k[0])
        * (
            1 / (k[2] - k[0]) * (np.exp(-k[0] * t) - np.exp(-k[2] * t))
            - 1 / (k[2] - k[1]) * (np.exp(-k[1] * t) - np.exp(-k[2] * t))
        )
    ),
}


def decay_get_Cref(k, y0, tout):
    coeffs = list(k) + [0] * (3 - len(k))

    return np.column_stack(
        [decay_analytic[i](y0, coeffs, tout) for i in range(min(3, len(k) + 1))]
    )


@requires("pycvodes", "pyodesys")
@pytest.mark.parametrize("solve", [(), ("CNO",)])
def test_get_native__first_step(solve):
    integrator = "cvode"
    forgive = 20

    def k(num):
        return "MassAction(unique_keys=('k%d',))" % num

    lines = [  # fictitious isomerization
        "CNO -> ONC; %s" % k(1),
        "ONC -> NCO; %s" % k(2),
        "NCO -> CON; %s" % k(3),
    ]
    rsys = ReactionSystem.from_string("\n".join(lines), "CNO ONC NCO CON")
    odesys, extra = get_odesys(rsys, include_params=False)
    if len(solve) > 0:
        from pyodesys.symbolic import PartiallySolvedSystem

        odesys = PartiallySolvedSystem(odesys, extra["linear_dependencies"](solve))
    c0 = defaultdict(float, {"CNO": 0.7})
    rate_coeffs = (1e78, 2, 3.0)
    args = (5, c0, dict(zip("k1 k2 k3".split(), rate_coeffs)))
    kwargs = dict(integrator=integrator, atol=1e-9, rtol=1e-9, nsteps=1000)
    native = get_native(rsys, odesys, integrator)

    h0 = extra["max_euler_step_cb"](0, *args[1:])
    xout1, yout1, info1 = odesys.integrate(*args, first_step=h0, **kwargs)
    xout2, yout2, info2 = native.integrate(*args, **kwargs)
    ref1 = decay_get_Cref(rate_coeffs, [c0[key] for key in native.names], xout1)
    ref2 = decay_get_Cref(rate_coeffs, [c0[key] for key in native.names], xout2)
    allclose_kw = dict(atol=kwargs["atol"] * forgive, rtol=kwargs["rtol"] * forgive)

    assert np.allclose(yout1[:, :3], ref1, **allclose_kw)

    assert info2["success"]
    assert (
        info2["nfev"] > 10
        and info2["nfev"] > 1
        and info2["time_cpu"] < 10
        and info2["time_wall"] < 10
    )
    assert np.allclose(yout2[:, :3], ref2, **allclose_kw)


@requires("pygslodeiv2", "pyodesys")
@pytest.mark.parametrize("solve", [(), ("H2O",)])
def test_get_native__a_substance_no_composition(solve):
    rsys = ReactionSystem.from_string(
        "\n".join(["H2O -> H2O+ + e-(aq); 1e-8", "e-(aq) + H2O+ -> H2O; 1e10"])
    )
    odesys, extra = get_odesys(rsys)
    c0 = {"H2O": 0, "H2O+": 2e-9, "e-(aq)": 3e-9}
    if len(solve) > 0:
        from pyodesys.symbolic import PartiallySolvedSystem

        odesys = PartiallySolvedSystem(odesys, extra["linear_dependencies"](solve))
    odesys = get_native(rsys, odesys, "gsl")
    xout, yout, info = odesys.integrate(1, c0, atol=1e-15, rtol=1e-15, integrator="gsl")
    c_reac = c0["H2O+"], c0["e-(aq)"]
    H2O_ref = binary_rev(xout, 1e10, 1e-4, c0["H2O"], max(c_reac), min(c_reac))
    assert np.allclose(yout[:, odesys.names.index("H2O")], H2O_ref)
    assert np.allclose(
        yout[:, odesys.names.index("H2O+")], c0["H2O+"] + c0["H2O"] - H2O_ref
    )
    assert np.allclose(
        yout[:, odesys.names.index("e-(aq)")], c0["e-(aq)"] + c0["H2O"] - H2O_ref
    )


@requires("pycvodes", "pyodesys")
@pytest.mark.parametrize("dep_scaling", [1, 763])
def test_get_native__named_parameter__units(dep_scaling):
    rsys = ReactionSystem.from_string(
        """
    -> H; 'p'
    H + H -> H2; 'k2'
    """,
        checks=("substance_keys", "duplicate", "duplicate_names"),
    )
    from pyodesys.symbolic import ScaledSys

    kwargs = (
        {} if dep_scaling == 1 else dict(SymbolicSys=ScaledSys, dep_scaling=dep_scaling)
    )
    odesys, extra = get_odesys(
        rsys, include_params=False, unit_registry=SI_base_registry, **kwargs
    )
    c0 = {"H": 42e-6 * u.molar, "H2": 17 * u.micromolar}
    native = get_native(rsys, odesys, "cvode")
    tend = 7 * 60 * u.minute
    g_rho_Ddot = g, rho, Ddot = 2 * u.per100eV, 998 * u.g / u.dm3, 314 * u.Gy / u.hour
    params = {"p": reduce(mul, g_rho_Ddot), "k2": 53 / u.molar / u.minute}
    result = native.integrate(
        tend, c0, params, atol=1e-15, rtol=1e-15, integrator="cvode", nsteps=16000
    )
    assert result.info["success"]

    def analytic_H(t, p, k, H0):
        # dH/dt = p - k2*H**2
        x0 = np.sqrt(2) * np.sqrt(p)
        x1 = x0
        x2 = np.sqrt(k)
        x3 = t * x1 * x2
        x4 = H0 * x2
        x5 = np.sqrt(x0 + 2 * x4)
        x6 = np.sqrt(-1 / (2 * H0 * x2 - x0))
        x7 = x5 * x6 * np.exp(x3)
        x8 = np.exp(-x3) / (x5 * x6)
        return x1 * (x7 - x8) / (2 * x2 * (x7 + x8))

    t_ul = to_unitless(result.xout, u.s)
    p_ul = to_unitless(params["p"], u.micromolar / u.s)
    ref_H_uM = analytic_H(
        t_ul,
        p_ul,
        to_unitless(params["k2"], 1 / u.micromolar / u.s),
        to_unitless(c0["H"], u.micromolar),
    )
    ref_H2_uM = (
        to_unitless(c0["H2"], u.micromolar)
        + to_unitless(c0["H"], u.micromolar) / 2
        + t_ul * p_ul / 2
        - ref_H_uM / 2
    )
    assert np.allclose(to_unitless(result.named_dep("H"), u.micromolar), ref_H_uM)
    assert np.allclose(to_unitless(result.named_dep("H2"), u.micromolar), ref_H2_uM)


@requires("pycvodes", "pyodesys")
def test_get_native__conc_roots():
    M, s = u.M, u.s
    rsys = ReactionSystem.from_string("2 O3 -> 3 O2; 'k2'")
    u_reg = SI_base_registry.copy()
    odesys, extra = get_odesys(rsys, include_params=False, unit_registry=u_reg)
    c0 = {"O3": 4.2e-3 * M, "O2": 0 * M}
    cr = ["O2"]
    native = get_native(rsys, odesys, "cvode", conc_roots=cr)
    tend = 1e5 * u.s
    params = {"k2": logspace_from_lin(1e-3 / M / s, 1e3 / M / s, 14)}
    tgt_O2 = 1e-3 * M
    results = native.integrate(
        tend,
        c0,
        params,
        integrator="native",
        return_on_root=True,
        special_settings=[unitless_in_registry(tgt_O2, u_reg)],
    )
    assert len(results) == params["k2"].size
    # dydt = -p*y**2
    # 1/y0 - 1/y = -2*pt
    # t = 1/2/p*(1/y - 1/y0)
    tgt_O3 = c0["O3"] - 2 / 3 * tgt_O2
    for r in results:
        ref = rescale(1 / 2 / r.named_param("k2") * (1 / tgt_O3 - 1 / c0["O3"]), u.s)
        assert allclose(r.xout[-1], ref, rtol=1e-6)


@requires("pycvodes", "pyodesys")
@pytest.mark.parametrize("scaling_density", [(1, False), (763, False), (1, True)])
def test_get_native__Radiolytic__named_parameter__units(scaling_density):
    scaling, density = scaling_density
    rsys = ReactionSystem.from_string(
        """
    -> H; Radiolytic(2*per100eV)
    H + H -> H2; 'k2'
    """,
        checks=("substance_keys", "duplicate", "duplicate_names"),
    )
    gval = 2 * u.per100eV

    from pyodesys.symbolic import ScaledSys

    kwargs = {} if scaling == 1 else dict(SymbolicSys=ScaledSys, dep_scaling=scaling)
    dens = {"density": 998 * u.g / u.dm3}
    odesys, extra = get_odesys(
        rsys,
        include_params=False,
        substitutions=dens if density else {},
        unit_registry=SI_base_registry,
        **kwargs
    )
    c0 = {"H": 42e-6 * u.molar, "H2": 17e3 * u.nanomolar}
    native = get_native(rsys, odesys, "cvode")
    tend = 7 * 60 * u.minute
    params = {"doserate": 314 * u.Gy / u.hour, "k2": 53 / u.molar / u.minute}
    if not density:
        params.update(dens)
    result = native.integrate(
        tend, c0, params, atol=1e-15, rtol=1e-15, integrator="cvode", nsteps=8000
    )
    assert result.info["success"]

    def analytic_H(t, p, k, H0):
        # dH/dt = p - k2*H**2
        x0 = np.sqrt(2) * np.sqrt(p)
        x1 = x0
        x2 = np.sqrt(k)
        x3 = t * x1 * x2
        x4 = H0 * x2
        x5 = np.sqrt(x0 + 2 * x4)
        x6 = np.sqrt(-1 / (2 * H0 * x2 - x0))
        x7 = x5 * x6 * np.exp(x3)
        x8 = np.exp(-x3) / (x5 * x6)
        return x1 * (x7 - x8) / (2 * x2 * (x7 + x8))

    t_ul = to_unitless(result.xout, u.s)
    p_ul = to_unitless(params["doserate"] * dens["density"] * gval, u.micromolar / u.s)
    ref_H_uM = analytic_H(
        t_ul,
        p_ul,
        to_unitless(params["k2"], 1 / u.micromolar / u.s),
        to_unitless(c0["H"], u.micromolar),
    )
    ref_H2_uM = (
        to_unitless(c0["H2"], u.micromolar)
        + to_unitless(c0["H"], u.micromolar) / 2
        + t_ul * p_ul / 2
        - ref_H_uM / 2
    )
    assert np.allclose(to_unitless(result.named_dep("H"), u.micromolar), ref_H_uM)
    assert np.allclose(to_unitless(result.named_dep("H2"), u.micromolar), ref_H2_uM)



from operator import itemgetter
from .printer import Printer
from ..units import is_quantity


class StrPrinter(Printer):
    _default_settings = dict(
        Printer._default_settings,
        repr_name="string",
        Equilibrium_arrow="=",
        Reaction_arrow="->",
    )

    def _Reaction_parts(self, rxn, **kwargs):
        str_ = self._str
        coeff_fmt = self._get("Reaction_coeff_fmt", **kwargs)
        formula_fmt = self._get("Reaction_formula_fmt", **kwargs)
        substances = self._get("substances", **kwargs) or {}
        nullstr, space = str_(""), str_(self._get("Reaction_coeff_space"))
        reac, prod, i_reac, i_prod = [
            [
                (
                    ((coeff_fmt(v) + space) if v != 1 else nullstr)
                    + formula_fmt(self._print(substances.get(k, k)))
                )
                for k, v in filter(itemgetter(1), d.items())
            ]
            for d in (rxn.reac, rxn.prod, rxn.inact_reac, rxn.inact_prod)
        ]
        r_str = str_(" + ").join(reac)
        ir_str = (
            str_(" + ( ") + str_(" + ").join(i_reac) + str_(")")
            if len(i_reac) > 0
            else nullstr
        )
        arrow_str = self._get("%s_arrow" % rxn.__class__.__name__, **kwargs)
        p_str = str_(" + ").join(prod)
        ip_str = (
            str_(" + ( ") + str_(" + ").join(i_prod) + str_(")")
            if len(i_prod) > 0
            else nullstr
        )
        return r_str, ir_str, arrow_str, p_str, ip_str

    def _Reaction_str(self, rxn, **kwargs):
        fmtstr = self._str("{}{}%s{}%s{}{}") % self._get(
            "Reaction_around_arrow", **kwargs
        )
        return fmtstr.format(*self._Reaction_parts(rxn, **kwargs))

    def _Reaction_param_str(self, rxn, **kwargs):
        mag_fmt = self._get("magnitude_fmt", **kwargs)
        unit_fmt = self._get("unit_fmt", **kwargs)
        try:
            magnitude_str = mag_fmt(rxn.param.magnitude)
            unit_str = unit_fmt(rxn.param.dimensionality)
        except AttributeError:
            if is_quantity(rxn.param) or isinstance(rxn.param, (float,)):
                return mag_fmt(rxn.param)
            else:
                return str(rxn.param)
        else:
            return magnitude_str + self._str(" ") + unit_str

    def _print_Reaction(self, rxn, **kwargs):
        res = self._Reaction_str(rxn, **kwargs)
        if self._get("with_param", **kwargs) and rxn.param is not None:
            res += self._get("Reaction_param_separator", **kwargs)
            try:
                res += getattr(rxn.param, self._get("repr_name", **kwargs))(
                    self._get("magnitude_fmt", **kwargs)
                )
            except AttributeError:
                res += self._Reaction_param_str(rxn, **kwargs)
        if self._get("with_name", **kwargs) and rxn.name is not None:
            res += self._get("Reaction_param_separator", **kwargs)
            res += rxn.name
        return res

    def _print_ReactionSystem(self, rsys, **kwargs):
        header = (rsys.name + "\n") if rsys.name else ""
        return header + "\n".join(map(self._print, rsys.rxns)) + "\n"


def str_(obj, **settings):  # Python keyword, hence the trailing '_'
    return StrPrinter(settings).doprint(obj)
